/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.objectweb.fractal.api.control.BindingController;

import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;

/**
 * A {@link Loader Loader} that loads type and template definitions from the
 * class path. This loader loads a definition named 'foo.bar.abc' by looking for
 * a file foo/bar/abc.<i>fractal</i> in the resources of the class loader of
 * this class (with the {@link ClassLoader#getResourceAsStream
 * getResourceAsStream} method). If such a file is found, this loader parses the
 * XML element it contains, which should be name 'foo.bar.abc', and returns it.
 */

public class BasicLoader implements Loader, BindingController {

  /**
   * The optional logger factory used to get a logger for this component.
   */

  protected LoggerFactory loggerFactory;

  /**
   * The logger used to log messages. May be <tt>null</tt>.
   */

  private Logger logger;

  // -------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // -------------------------------------------------------------------------

  public String[] listFc () {
    return new String[] { "logger-factory" };
  }

  public Object lookupFc (final String clientItfName) {
    if (clientItfName.equals("logger-factory")) {
      return loggerFactory;
    }
    return null;
  }

  public void bindFc (final String clientItfName, final Object serverItf) {
    if (clientItfName.equals("logger-factory")) {
      loggerFactory = (LoggerFactory)serverItf;
      logger = loggerFactory.getLogger(getClass().getName());
    }
  }

  public void unbindFc (final String clientItfName) {
    if (clientItfName.equals("logger-factory")) {
      loggerFactory = null;
      logger = null;
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the Loader interface
  // -------------------------------------------------------------------------

  public XMLElement loadTypeDescriptor (final String name)
    throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "loading type descriptor: " + name);
    }
    return load(name);
  }

  public XMLElement loadTemplateDescriptor (final String name)
    throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "loading template descriptor: " + name);
    }
    return load(name);
  }

  // -------------------------------------------------------------------------
  // Utility methods
  // -------------------------------------------------------------------------

  /**
   * Loads the definition whose name is given.
   *
   * @param name the name of type or template definition.
   * @return the XML element corresponding to the specified definition.
   * @throws ParserException if the definition cannot be found.
   */

  private XMLElement load (final String name) throws ParserException {
    String file = name.replace('.', '/') + ".fractal";
    ClassLoader cl = getClass().getClassLoader();
    if (cl == null) {
      cl = ClassLoader.getSystemClassLoader();
    }
    InputStream is = cl.getResourceAsStream(file);
    if (is == null) {
      throw new ParserException(
        "Cannot find '" + file + "' in the classpath", null);
    }
    String sourceComment = null;
    XMLElement xml;
    try {
      LineNumberReader lnr = new LineNumberReader(new InputStreamReader(is));
      do {
        xml = XMLElement.parse(lnr, true);
        if (xml.tag.startsWith("<!-- copied from ")) {
          sourceComment = xml.tag;
        }
      } while (xml.tag.startsWith("<!"));
    } catch (IOException ioe) {
      throw new ParserException("File " + file, ioe);
    } catch (ParserException e) {
      throw new ParserException(
        file, e.getLine(), e.getMessage(), e.getException());
    }
    if (!xml.getAttribute("name", "").equals(name)) {
      throw new ParserException(
        file,
        xml.line,
        "This element should be named '" + name + "'",
        null);
    }
    String source;
    int offset;
    try {
      sourceComment = sourceComment.substring(17, sourceComment.length() - 3);
      sourceComment = sourceComment.trim();
      int p = sourceComment.lastIndexOf(':');
      source = sourceComment.substring(0, p);
      offset = Integer.parseInt(sourceComment.substring(p + 1)) - xml.line;
    } catch (Exception e) {
      source = file;
      offset = 0;
    }
    updateSource(xml, source, offset);
    return xml;
  }

  /**
   * Sets the source and line fields of the given XML element and of all its
   * sub elements.
   *
   * @param elem an XML element.
   * @param source a source file name.
   * @param offset a line number offset.
   */

  private void updateSource (
    final XMLElement elem,
    final String source,
    final int offset)
  {
    elem.source = source;
    elem.line += offset;
    for (int i = 0; i < elem.getSubElementCount(); ++i) {
      updateSource(elem.getSubElement(i), source, offset);
    }
  }
}
