/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.control.AttributeController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.factory.GenericFactory;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.api.type.TypeFactory;

import org.objectweb.fractal.util.Fractal;

import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

/**
 * Default implementation of the {@link Parser Parser} interface.
 */

public class BasicParser implements Parser, BindingController {

  /**
   * A cache of already loaded types and templates.
   */

  private HashMap cache = new HashMap();

  /**
   * A map associating external components to their names.
   */

  private HashMap externals = new HashMap();

  /**
   * The loader used to load resolved type and template definitions.
   */

  protected Loader loader;

  /**
   * The factory used to create the component types.
   */

  protected TypeFactory typeFactory;

  /**
   * The factory used to create the component templates.
   */

  protected GenericFactory genericFactory;

  /**
   * The optional logger factory used to get a logger for this component.
   */

  protected LoggerFactory loggerFactory;

  /**
   * The logger used to log messages. May be <tt>null</tt>.
   */

  private Logger logger;

  public BasicParser () {
    try {
      Component boot = Fractal.getBootstrapComponent();
      externals.put("org.objectweb.fractal.api.bootstrap", boot);
    } catch (Exception ignored) {
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // -------------------------------------------------------------------------

  public String[] listFc () {
    return new String[] {
      "loader", "type-factory", "generic-factory", "logger-factory"
    };
  }

  public Object lookupFc (final String clientItfName) {
    if (clientItfName.equals("loader")) {
      return loader;
    } else if (clientItfName.equals("type-factory")) {
      return typeFactory;
    } else if (clientItfName.equals("generic-factory")) {
      return genericFactory;
    } else if (clientItfName.equals("logger-factory")) {
      return loggerFactory;
    }
    return null;
  }

  public void bindFc (final String clientItfName, final Object serverItf) {
    if (clientItfName.equals("loader")) {
      loader = (Loader)serverItf;
    } else if (clientItfName.equals("type-factory")) {
      typeFactory = (TypeFactory)serverItf;
    } else if (clientItfName.equals("generic-factory")) {
      genericFactory = (GenericFactory)serverItf;
    } else if (clientItfName.equals("logger-factory")) {
      loggerFactory = (LoggerFactory)serverItf;
      logger = loggerFactory.getLogger(getClass().getName());
    }
  }

  public void unbindFc (final String clientItfName) {
    if (clientItfName.equals("loader")) {
      loader = null;
    } else if (clientItfName.equals("type-factory")) {
      typeFactory = null;
    } else if (clientItfName.equals("generic-factory")) {
      genericFactory = null;
    } else if (clientItfName.equals("logger-factory")) {
      loggerFactory = null;
      logger = null;
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the TemplateFactory interface
  // -------------------------------------------------------------------------

  public ComponentType loadType (final String name)
    throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "parsing type descriptor: " + name);
    }
    ComponentType result = (ComponentType)cache.get(name);
    if (result == null) {
      XMLElement desc;
      try {
        desc = loader.loadTypeDescriptor(name);
      } catch (ParserException e) {
        throw new ParserException("Cannot load type '" + name + "'", e);
      }
      try {
        result = createComponentType(desc);
      } catch (ParserException e) {
        throw new ParserException(desc, "Cannot create type '" + name + "'",
          new ParserException(desc, e));
      }
      cache.put(name, result);
    }
    return result;
  }

  public Component loadTemplate (final String name, final boolean template)
    throws ParserException
  {
    String key = (template ? "T-" : "C-") + name;
    Component result = (Component)cache.get(key);
    if (result == null) {
      result = createTemplate(name, template);
      cache.put(key, result);
    }
    return result;
  }

  public Component createTemplate (final String name, final boolean template)
    throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "parsing template descriptor: " + name);
    }
    XMLElement desc;
    try {
      desc = loader.loadTemplateDescriptor(name);
    } catch (ParserException e) {
      throw new ParserException("Cannot load template '" + name + "'", e);
    }
    HashMap infos = new HashMap();
    // first pass: analyzes the dependencies between templates
    analyzeTemplate(name, desc, new HashMap(), infos);
    // second pass: creates the templates
    createTemplates(infos, template);
    // return result
    return ((TemplateInfo)infos.get(name)).value;
  }

  public void registerExternalComponent (
    final String name,
    final Component component)
  {
    externals.put(name, component);
  }

  // -------------------------------------------------------------------------
  // Utility methods: component types
  // -------------------------------------------------------------------------

  /**
   * Creates the component type corresponding to the given description.
   *
   * @param node a &lt;component-type&gt; XML element.
   * @return the component type corresponding to the given description.
   * @throws ParserException if the component type cannot be created.
   */

  private ComponentType createComponentType (final XMLElement node)
    throws ParserException
  {
    ArrayList itfList = new ArrayList();
    for (int i = 0; i < node.getSubElementCount(); ++i) {
      XMLElement subElement = node.getSubElement(i);
      createInterfaceTypes(subElement, itfList);
    }
    InterfaceType[] itfTypes = new InterfaceType[itfList.size()];
    itfTypes = (InterfaceType[])itfList.toArray(itfTypes);
    try {
      return typeFactory.createFcType(itfTypes);
    } catch (Exception e) {
      throw new ParserException(node, "Cannot create component type", e);
    }
  }

  /**
   * Creates the interface types corresponding to the given descriptions.
   *
   * @param node a &lt;provides&gt; or &lt;requires&gt; XML element.
   * @param itfList the list to which the created interface types must be added.
   * @throws ParserException if an interface type cannot be create.
   */

  private void createInterfaceTypes (
    final XMLElement node,
    final ArrayList itfList) throws ParserException
  {
    boolean server = node.tag.equals("provides");
    for (int i = 0; i < node.getSubElementCount(); ++i) {
      XMLElement subElement = node.getSubElement(i);
      String prefix = subElement.getAttribute("name", null);
      String type = subElement.getAttribute("signature", null);
      boolean mandatory =
        subElement.getAttribute("contingency", "mandatory").equals("mandatory");
      boolean singleton =
        subElement.getAttribute("cardinality", "single").equals("single");
      if (!singleton) {
        mandatory = false;
      }
      try {
        itfList.add(typeFactory.createFcItfType(
          prefix, type, !server, !mandatory, !singleton));
      } catch (Exception e) {
        throw new ParserException(node, "Cannot create interface type", e);
      }
    }
  }

  // -------------------------------------------------------------------------
  // Utility methods: general template creation algorithm
  // -------------------------------------------------------------------------

  /**
   * Analyzes the given template definition. This recursive method analyzes a
   * template definition and all its sub templates, and puts the collected
   * dependency information in the given hash map.
   *
   * @param path the path of the given template from the root template.
   * @param tmpl the template definition to be analyzed.
   * @param shared information that has been collected about shared components.
   *      This map contains the paths, from the root template, corresponding to
   *      the content of the &lt;sharing&gt; elements that have been analyzed.
   *      Each entry of this map associates a &lt;ref&gt; to a &lt;path&gt;.
   * @param infos information that has been collected about the templates that
   *      been analyzed. Each entry of this map associates a
   *      <tt>TemplateInfo</tt> to a template, designated by its path from the
   *      root template.
   * @throws ParserException if a problem occurs during the analysis.
   */

  private void analyzeTemplate (
    final String path,
    final XMLElement tmpl,
    final HashMap shared,
    final HashMap infos) throws ParserException
  {
    try {
      TemplateInfo info = new TemplateInfo();
      info.descriptor = tmpl;
      infos.put(path, info);
      if (tmpl.tag.equals("composite-template")) {
        XMLElement content = tmpl.getSubElement("composite-content");
        if (content == null) {
          return;
        }
        XMLElement sharing = content.getSubElement("sharing");
        if (sharing != null) {
          for (int i = 0; i < sharing.getSubElementCount(); ++i) {
            XMLElement elem = sharing.getSubElement(i);
            String epath = elem.getAttribute("path", null);
            String eref = elem.getAttribute("ref", null);
            shared.put(path + '/' + epath, path + '/' + eref);
          }
        }
        XMLElement comps = content.getSubElement("components");
        if (comps == null) {
          return;
        }
        info.requires = new String[comps.getSubElementCount()];
        for (int i = 0; i < info.requires.length; ++i) {
          XMLElement comp = comps.getSubElement(i);
          String name = path + '/' + comp.getAttribute("name", null);
          String ref = (String)shared.get(name);
          if (ref != null) {
            while (ref != null) {
              info.requires[i] = ref;
              ref = (String)shared.get(ref);
            }
          } else {
            info.requires[i] = name;
            String subName;
            try {
              subName = comp.getAttribute("implementation", null);
            } catch (ParserException e) {
              // optional component, whose implementation is not provided
              continue;
            }
            XMLElement subTmpl;
            try {
              subTmpl = loader.loadTemplateDescriptor(subName);
            } catch (ParserException e) {
              throw new ParserException(
                comp, "Cannot load template '" + subName + "'", e);
            }
            analyzeTemplate(name, subTmpl, shared, infos);
          }
        }
      }
    } catch (ParserException e) {
      throw new ParserException(tmpl, "Error while analyzing '" + path + "'",
        new ParserException(tmpl, e));
    }
  }

  /**
   * Creates the template components corresponding to the given definitions.
   * This method creates all the templates that do not have any dependencies,
   * then the templates that depend only on the previously instantiated
   * templates, and so on until all templates have been instantiated (or a
   * dependency cycle is detected, in which case an exception is thrown).
   *
   * @param infos information about the templates that must be created. Each
   *      entry of this map associates a <tt>TemplateInfo</tt> to a template,
   *      defined by its path from the root template.
   * @param template <tt>true</tt> to return a template components,
   *      <tt>false</tt> to return directly the instantiated components.
   * @throws ParserException if a problem occurs during the creation of the
   *      template components.
   */

  private void createTemplates (final HashMap infos, final boolean template)
    throws ParserException
  {
    HashMap map = new HashMap(infos);
    while (map.size() > 0) {
      boolean ok = false;
      Iterator i = map.keySet().iterator();
      while (i.hasNext()) {
        String path = (String)i.next();
        TemplateInfo info = (TemplateInfo)map.get(path);
        if (info.value == null && info.instantiable(infos)) {
          createTemplate(info, path, infos, template);
          i.remove();
          ok = true;
        }
      }
      if (!ok) {
        throw new ParserException(
          "Dependency cycle due to shared components detected", null);
      }
    }
  }

  /**
   * Creates a template component corresponding to the given definition. This
   * method is only called when all the sub templates needed to create the
   * given template have already been created. These required sub templates
   * are retrieved from the given hash map.
   *
   * @param info information about the template component to be created.
   * @param path path of this template from the root template.
   * @param infos information about already created templates. Each entry
   *      of this map associates a <tt>TemplateInfo</tt> to a template,
   *      defined by its path from the root template.
   * @param template <tt>true</tt> to return a template component,
   *      <tt>false</tt> to return directly the instantiated component.
   * @throws ParserException if a problem occurs during the creation of the
   *      template component.
   */

  private void createTemplate (
    final TemplateInfo info,
    final String path,
    final HashMap infos,
    final boolean template) throws ParserException
  {
    XMLElement tmpl = info.descriptor;
    try {
      if (tmpl.tag.equals("primitive-template")) {
        info.value = createPrimitiveTemplate(tmpl, template);
      } else if (tmpl.tag.equals("composite-template")) {
        info.value = createCompositeTemplate(info, infos, template);
      } else {
        info.value = createExternalTemplate(info, template);
      }
    } catch (ParserException e) {
      throw new ParserException(tmpl, "Error while creating '" + path + "'",
        new ParserException(tmpl, e));
    }
  }

  // -------------------------------------------------------------------------
  // Utility methods: primitive template creation
  // -------------------------------------------------------------------------

  /**
   * Creates a primitive template corresponding to the given description.
   *
   * @param tmpl a &lt;primitive-template&gt; XML element.
   * @param template <tt>true</tt> to return a template component,
   *      <tt>false</tt> to return directly the instantiated component.
   * @return a primitive template corresponding to the given description.
   * @throws ParserException if the template component cannot be created.
   */

  private Component createPrimitiveTemplate (
    final XMLElement tmpl,
    final boolean template) throws ParserException
  {
    ComponentType type = loadType(tmpl.getAttribute("implements", null));
    XMLElement content = tmpl.getSubElement("primitive-content");
    XMLElement controller = tmpl.getSubElement("controller");
    String tmplDesc = "primitiveTemplate";
    String compDesc = "primitive";
    XMLElement attrs = null;
    if (controller != null) {
      for (int i = 0; i < controller.getSubElementCount(); ++i) {
        XMLElement elem = controller.getSubElement(i);
        if (elem.tag.equals("attributes")) {
          attrs = elem;
          tmplDesc = "parametricPrimitiveTemplate";
        }
      }
      for (int i = 0; i < controller.getSubElementCount(); ++i) {
        XMLElement elem = controller.getSubElement(i);
        if (elem.tag.equals("template-controller")) {
          tmplDesc = elem.getAttribute("desc", null);
        } else if (elem.tag.equals("component-controller")) {
          compDesc = elem.getAttribute("desc", null);
        }
      }
    }
    if (attrs != null) {
      type = addAttributeController(type, attrs);
    }
    Component tmplId;
    try {
      if (template) {
        tmplId = genericFactory.newFcInstance(
          type, tmplDesc, new Object[] {
            compDesc, content.getAttribute("class", null)
          });
      } else {
        tmplId = genericFactory.newFcInstance(
          type, compDesc, content.getAttribute("class", null));
      }
    } catch (Exception e) {
      throw new ParserException(
        tmpl, "Cannot create template component", e);
    }
    if (attrs != null) {
      setAttributes(attrs, tmplId);
    }
    setName(tmplId, tmpl.getAttribute("name", null));
    return tmplId;
  }

  /**
   * Adds an attribute controller interface type to the given component type.
   *
   * @param type a component type.
   * @param attrs an &lt;attributes&gt; XML element.
   * @return the modified component type.
   * @throws ParserException if a problem occurs.
   */

	private ComponentType addAttributeController (
    final ComponentType type,
    final XMLElement attrs) throws ParserException
  {
    InterfaceType[] itfTypes = type.getFcInterfaceTypes();
    InterfaceType[] newItfTypes = new InterfaceType[itfTypes.length + 1];
    System.arraycopy(itfTypes, 0, newItfTypes, 1, itfTypes.length);
    try {
      newItfTypes[0] = typeFactory.createFcItfType(
        "attribute-controller",
        attrs.getAttribute("signature", null),
        false,
        false,
        false);
    } catch (Exception e) {
      throw new ParserException(
        attrs, "Cannot create attribute controller interface type", e);
    }
    try {
      return typeFactory.createFcType(newItfTypes);
    } catch (Exception e) {
      throw new ParserException(
        attrs, "Cannot create component type", e);
    }
  }

  /**
   * Sets the values of the specified attributes on the given template.
   *
   * @param attrs an &lt;attributes&gt; XML element.
   * @param tmpl a template component.
   * @throws ParserException if the value of an attribute cannot be set.
   */

  private void setAttributes (
    final XMLElement attrs,
    final Component tmpl) throws ParserException
  {
    AttributeController attrCont;
    try {
      attrCont = (AttributeController)tmpl.getFcInterface(
        "attribute-controller");
    } catch (Exception e) {
      throw new ParserException(
        attrs, "Cannot get attribute controller interface", e);
    }
    Class attrContClass = attrCont.getClass();
    for (int i = 0; i < attrs.getSubElementCount(); ++i) {
      XMLElement attr = attrs.getSubElement(i);
      String attrName = attr.getAttribute("name", null);
      String attrValue = attr.getAttribute("value", null);
      // sets the attribute's value
      try {
        String getterName = "get" + attrName;
        String setterName = "set" + attrName;
        Method getter = attrContClass.getMethod(getterName, new Class[0]);
        Method setter = attrContClass.getMethod(setterName, new Class[] {
          getter.getReturnType()
        });
        Class attrType = getter.getReturnType();
        Object value;
        if (attrType.equals(String.class)) {
          value = attrValue;
        } else if (attrType.isPrimitive()) {
          if (attrType.equals(Integer.TYPE)) {
            value = Integer.valueOf(attrValue);
          } else if (attrType.equals(Long.TYPE)) {
            value = Long.valueOf(attrValue);
          } else if (attrType.equals(Float.TYPE)) {
            value = Float.valueOf(attrValue);
          } else if (attrType.equals(Double.TYPE)) {
            value = Double.valueOf(attrValue);
          } else if (attrType.equals(Byte.TYPE)) {
            value = Byte.valueOf(attrValue);
          } else if (attrType.equals(Character.TYPE)) {
            if (attrValue.length() != 1) {
              throw new Exception("Bad char value: " + attrValue);
            }
            value = new Character(attrValue.charAt(0));
          } else if (attrType.equals(Short.TYPE)) {
            value = Short.valueOf(attrValue);
          } else if (attrType.equals(Boolean.TYPE)) {
            if (!attrValue.equals("true") &&
                !attrValue.equals("false"))
            {
              throw new Exception("Bad boolean value: " + attrValue);
            }
            value = new Boolean(attrValue.equals("true"));
          } else {
            throw new Exception("Unexpected case");
          }
        } else {
          throw new Exception("Unsupported attribute type: " + attrType);
        }
        setter.invoke(attrCont, new Object[]{value});
      } catch (Exception e) {
        throw new ParserException(
          attr, "Cannot set '" + attrName + "' attribute value", e);
      }
    }
  }

  // -------------------------------------------------------------------------
  // Utility methods: composite template creation
  // -------------------------------------------------------------------------

  /**
   * Creates a composite template corresponding to the given definition. This
   * method is only called when all the sub templates needed to create the
   * given template have already been created. These required sub templates
   * are retrieved from the given hash map.
   *
   * @param info information about the template component to be created.
   * @param infos information about already created templates. Each entry
   *      of this map associates a <tt>TemplateInfo</tt> to a template,
   *      defined by its path from the root template.
   * @return the Component interface of the created template.
   * @throws ParserException if a problem occurs during the creation of the
   *      template component.
   */

  private Component createCompositeTemplate (
    final TemplateInfo info,
    final HashMap infos,
    final boolean template) throws ParserException
  {
    XMLElement tmpl = info.descriptor;
    XMLElement content = tmpl.getSubElement("composite-content");
    XMLElement controller = tmpl.getSubElement("controller");
    XMLElement components = null;
    XMLElement attrs = null;
    if (content != null) {
      components = content.getSubElement("components");
    }

    ComponentType type = loadType(tmpl.getAttribute("implements", null));
    String tmplDesc = "compositeTemplate";
    String compDesc = "composite";

    if (controller != null) {
      for (int i = 0; i < controller.getSubElementCount(); ++i) {
        XMLElement elem = controller.getSubElement(i);
        if (elem.tag.equals("attributes")) {
          attrs = elem;
        } else if (elem.tag.equals("template-controller")) {
          tmplDesc = elem.getAttribute("desc", null);
        } else if (elem.tag.equals("component-controller")) {
          compDesc = elem.getAttribute("desc", null);
        }
      }
    }
    if (attrs != null) {
      type = addAttributeController(type, attrs);
    }
    // creates the composite template
    Component tmplId;
    try {
      if (template) {
        tmplId = genericFactory.newFcInstance(
          type, tmplDesc, new Object[] {compDesc, null});
      } else {
        tmplId = genericFactory.newFcInstance(
          type, compDesc, null);
      }
    } catch (Exception e) {
      throw new ParserException(tmpl, "Cannot create template component", e);
    }
    if (attrs != null) {
      setAttributes(attrs, tmplId);
    }
    setName(tmplId, tmpl.getAttribute("name", null));

    // adds its sub template components
    ContentController cc;
    try {
      cc = Fractal.getContentController(tmplId);
    } catch (Exception e) {
      throw new ParserException(
        tmpl, "Cannot get the template's content controller interface", e);
    }
    HashMap subComps = new HashMap();
    subComps.put("this", tmplId);
    int n = info.requires == null ? 0 : info.requires.length;
    for (int i = 0; i < n; ++i) {
      String path = info.requires[i];
      String name = components.getSubElement(i).getAttribute("name", null);
      TemplateInfo tmplInfo = (TemplateInfo)infos.get(path);
      if (tmplInfo == null) {
        // optional component, whose implementation is not provided
        continue;
      }
      Component id = tmplInfo.value;
      setName(id, name);
      subComps.put(name, id);
      try {
        cc.addFcSubComponent(id);
      } catch (Exception e) {
        throw new ParserException(
          tmpl, "Cannot add the '" + name + "' sub template", e);
      }
    }

    // creates the bindings between the sub template components
    if (content == null) {
      return tmplId;
    }
    XMLElement bindings = content.getSubElement("bindings");
    if (bindings == null) {
      return tmplId;
    }
    for (int i = 0; i < bindings.getSubElementCount(); ++i) {
      XMLElement bind = bindings.getSubElement(i);
      String clt = bind.getAttribute("client", null);
      int p = clt.indexOf('.');
      String cltComp = clt.substring(0, p);
      String cltItf = clt.substring(p + 1);
      String srv = bind.getAttribute("server", null);
      p = srv.indexOf('.');
      String srvComp = srv.substring(0, p);
      String srvItf = srv.substring(p + 1);

      Component client = (Component)subComps.get(cltComp);
      Component server = (Component)subComps.get(srvComp);

      if (client == null || server == null) {
        // one of the component is an optional component which is not provided:
        // skips this binding
        continue;
      }
      if (client == null) {
        throw new ParserException(
          bind, "No such sub template: " + cltComp, null);
      }
      if (server == null) {
        throw new ParserException(
          bind, "No such sub template: " + srvComp, null);
      }
      Interface sItf;
      try {
        sItf = (Interface)server.getFcInterface(srvItf);
        if (((InterfaceType)sItf.getFcItfType()).isFcClientItf()) {
          cc = Fractal.getContentController(server);
          sItf = (Interface)cc.getFcInternalInterface(srvItf);
        }
      } catch (Exception e) {
        throw new ParserException(bind, "No such interface: " + srv, e);
      }
      String cltItfPrefix = null;
      type = (ComponentType)client.getFcType();
      InterfaceType[] itfTypes = type.getFcInterfaceTypes();
      for (int j = 0; j < itfTypes.length; ++j) {
        if (itfTypes[j].isFcCollectionItf()) {
          if (cltItf.startsWith(itfTypes[j].getFcItfName())) {
            cltItfPrefix = itfTypes[j].getFcItfName();
            break;
          }
        } else {
          if (cltItf.equals(itfTypes[j].getFcItfName())) {
            cltItfPrefix = itfTypes[j].getFcItfName();
            break;
          }
        }
      }
      if (cltItfPrefix == null) {
        throw new ParserException(bind, "No such interface: " + clt, null);
      }
      try {
        Fractal.getBindingController(client).bindFc(cltItf, sItf);
      } catch (Exception e) {
        throw new ParserException(bind, "Cannot create binding", e);
      }
    }
    return tmplId;
  }

  // -------------------------------------------------------------------------
  // Utility methods: external template creation
  // -------------------------------------------------------------------------

  /**
   * Creates an external template corresponding to the given description.
   *
   * @param info information about the template component to be created.
   * @param template <tt>true</tt> to return a template component,
   *      <tt>false</tt> to return directly the instantiated component.
   * @return a template corresponding to the given description.
   * @throws ParserException if the template component cannot be created.
   */

  private Component createExternalTemplate (
    final TemplateInfo info,
    final boolean template) throws ParserException
  {
    XMLElement ref = info.descriptor.getSubElement(0);
    String id = ref.getAttribute("name", null);
    Component comp = (Component)externals.get(id);
    if (comp == null) {
      throw new ParserException(ref, "No such external component: " + id, null);
    }
    if (!template) {
      return comp;
    }
    try {
      return genericFactory.newFcInstance(null, "singletonTemplate", new Object[] {null, comp});
    } catch (Exception e) {
      e.printStackTrace();
      throw new ParserException(
        info.descriptor, "Cannot create external template", e);
    }
  }

  /**
   * Sets the name of the given component, if it has a <tt>NameController</tt>
   * interface.
   *
   * @param c the Component interface of a component.
   * @param name the name this component must have.
   */

  private static void setName (final Component c, final String name) {
    try {
      NameController nc = Fractal.getNameController(c);
      nc.setFcName(name);
    } catch (Exception ignored) {
    }
  }

  // -------------------------------------------------------------------------
  // Utility class
  // -------------------------------------------------------------------------

  /**
   * Information about a template.
   */

  static class TemplateInfo {

    /**
     * A template definition.
     */

    XMLElement descriptor;

    /**
     * The paths of the sub templates that are required to create the template
     * defined by <tt>descriptor</tt>. May be <tt>null</tt> if there are no such
     * sub templates.
     */

    String[] requires;

    /**
     * The template component corresponding to <tt>descriptor</tt>. May be
     * <tt>null</tt> if this component has not been created yet.
     */

    Component value;

    /**
     * Tests if the template defined by <tt>descriptor</tt> can be instantiated.
     *
     * @param infos information about already instantiated templates. Each entry
     *      of this map associates a <tt>TemplateInfo</tt> to a template,
     *      defined by its path from the root template.
     * @return <tt>true</tt> if the template defined by <tt>descriptor</tt> can
     *      be instantiated, i.e., if all the templates it requires have already
     *      been instantiated (i.e., if the <tt>value</tt> field of their
     *      associated <tt>TemplateInfo</tt> object is not <tt>null</tt>).
     */

    public boolean instantiable (final HashMap infos) {
      if (requires != null) {
        for (int i = 0; i < requires.length; ++i) {
          TemplateInfo tmplInfo = (TemplateInfo)infos.get(requires[i]);
          if (tmplInfo != null && tmplInfo.value == null) {
            return false;
          }
        }
      }
      return true;
    }
  }
}
