/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.objectweb.util.monolog.api.BasicLevel;

import java.lang.reflect.Method;

/**
 * A {@link CachingLoader CachingLoader} that checks typing constraints on
 * the elements it loads.
 */

public class CheckingLoader extends CachingLoader {

  // -------------------------------------------------------------------------
  // Implementation of the abstract inherited methods
  // -------------------------------------------------------------------------

  protected XMLElement handleTypeDescriptor (
    final String name,
    final XMLElement desc) throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "checking type descriptor: " + name);
    }
    checkComponentType(desc);
    return desc;
  }

  protected XMLElement handleTemplateDescriptor
    (final String name,
    final XMLElement desc) throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "checking template descriptor: " + name);
    }
    loadTypeDescriptor(desc.getAttribute("implements", null));
    if (desc.tag.equals("primitive-template")) {
      checkPrimitiveTemplate(desc);
    } else if (desc.tag.equals("composite-template")) {
      checkCompositeTemplate(desc);
    } else {
      checkExternalTemplate(desc);
    }
    return desc;
  }

  // -------------------------------------------------------------------------
  // Utility methods
  // -------------------------------------------------------------------------

  // component types ---------------------------------------------------------

  /**
   * Checks a 'component-type' element.
   *
   * @param type the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkComponentType (final XMLElement type)
    throws ParserException
  {
    for (int i = 0; i < type.getSubElementCount(); ++i) {
      XMLElement elem = type.getSubElement(i);
      checkInterfaceTypes(elem);
    }
  }

  /**
   * Checks a 'provides' or 'requires' element.
   *
   * @param types the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkInterfaceTypes (final XMLElement types)
    throws ParserException
  {
    for (int i = 0; i < types.getSubElementCount(); ++i) {
      XMLElement elem = types.getSubElement(i);
      checkInterfaceType(elem);
    }
  }

  /**
   * Checks an 'interface-type' element.
   *
   * @param type the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkInterfaceType (final XMLElement type)
    throws ParserException
  {
    // checks that the 'signature' Java interface exists
    String s = type.getAttribute("signature", null);
    try {
      Class.forName(s);
    } catch (ClassNotFoundException e) {
      throw new ParserException(type, "Class not found: " + s, e);
    }
  }

  // primitive templates -----------------------------------------------------

  /**
   * Checks a 'primitive-template' element.
   *
   * @param tmpl the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkPrimitiveTemplate (final XMLElement tmpl)
    throws ParserException
  {
    loadTypeDescriptor(tmpl.getAttribute("implements", null));
    for (int i = 0; i < tmpl.getSubElementCount(); ++i) {
      XMLElement elem = tmpl.getSubElement(i);
      if (elem.tag.equals("primitive-content")) {
        checkPrimitiveContent(elem);
      } else {
        checkController(elem);
      }
    }
  }

  /**
   * Checks a 'primitive-content' element.
   *
   * @param content the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkPrimitiveContent (final XMLElement content)
    throws ParserException
  {
    // checks that the 'content' Java class exists
    String s = content.getAttribute("class", null);
    try {
      Class.forName(s);
    } catch (ClassNotFoundException e) {
      throw new ParserException(content, "Class not found: " + s, e);
    }
  }

  // composite templates -----------------------------------------------------

  /**
   * Checks a 'composite-template' element.
   *
   * @param tmpl the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkCompositeTemplate (final XMLElement tmpl)
    throws ParserException
  {
    String name = tmpl.getAttribute("name", null);
    String type = tmpl.getAttribute("implements", null);
    loadTypeDescriptor(type);
    for (int i = 0; i < tmpl.getSubElementCount(); ++i) {
      XMLElement elem = tmpl.getSubElement(i);
      if (elem.tag.equals("composite-content")) {
        checkCompositeContent(name, type, elem);
      } else {
        checkController(elem);
      }
    }
  }

  /**
   * Checks a 'composite-content' element.
   *
   * @param name the name of the current composite template.
   * @param type the type name of the current composite template.
   * @param content the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkCompositeContent (
    final String name,
    final String type,
    final XMLElement content) throws ParserException
  {
    XMLElement comps = null;
    for (int i = 0; i < content.getSubElementCount(); ++i) {
      XMLElement elem = content.getSubElement(i);
      if (elem.tag.equals("components")) {
        checkComponents(elem);
        comps = elem;
      } else if (elem.tag.equals("bindings")) {
        checkBindings(type, comps, elem);
      } else if (elem.tag.equals("sharing")) {
        checkSharing(name, elem);
      }
    }
  }

  /**
   * Checks a 'components' element.
   *
   * @param comps the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkComponents (final XMLElement comps)
    throws ParserException
  {
    for (int i = 0; i < comps.getSubElementCount(); ++i) {
      XMLElement elem = comps.getSubElement(i);
      checkComponent(elem);
    }
  }

  /**
   * Checks a 'component' sub element of a 'components' element.
   *
   * @param comp the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkComponent (final XMLElement comp)
    throws ParserException
  {
    // checks that the 'implementation' component's type is a sub type of 'type'
    String impl;
    try {
      impl = comp.getAttribute("implementation", null);
    } catch (ParserException e) {
      return;
    }
    XMLElement tmpl;
    try {
      tmpl = loadTemplateDescriptor(impl);
    } catch (ParserException e) {
      throw new ParserException(comp, "Cannot load '" + impl + '\'', e);
    }
    String rType = comp.getAttribute("type", null);
    String pType = tmpl.getAttribute("implements", null);
    try {
      checkSubType(pType, rType);
    } catch (ParserException e) {
      throw new ParserException(
        comp,
        "The type of '" + impl + "' is not a subtype of '" + rType + "'",
        e);
    }
  }

  /**
   * Checks a 'bindings' element.
   *
   * @param type the type name of the current composite template.
   * @param comps the &lt;components&gt; element of the current composite
   *      template.
   * @param bindings the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkBindings (
    final String type,
    final XMLElement comps,
    final XMLElement bindings) throws ParserException
  {
    for (int i = 0; i < bindings.getSubElementCount(); ++i) {
      XMLElement elem = bindings.getSubElement(i);
      checkBinding(type, comps, elem);
    }
  }

  /**
   * Checks a 'binding' element.
   *
   * @param type the type name of the current composite template.
   * @param comps the &lt;components&gt; element of the current composite
   *      template.
   * @param binding the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkBinding (
    final String type,
    final XMLElement comps,
    final XMLElement binding) throws ParserException
  {
    // checks that the client and server components exist,
    // checks that the client and server interfaces exist,
    // and checks that the two interfaces have compatible types.
    String s = binding.getAttribute("client", null);
    XMLElement[] itfTypes = new XMLElement[2];
    for (int i = 0; i < 2; ++i) {
      int p = s.indexOf('.');
      String comp = s.substring(0, p);
      String itf = s.substring(p + 1);
      String typ;
      if (comp.equals("this")) {
        typ = type;
      } else {
        XMLElement compElem = findComponent(comps, comp);
        if (compElem == null) {
          throw new ParserException(binding, "No such component: " + comp, null);
        }
        typ = compElem.getAttribute("type", null);
      }
      if (!itf.equals("component") && !itf.endsWith("-controller")) {
        boolean provided = comp.equals("this") == (i == 0);
        itfTypes[i] = findInterface(typ, itf, provided);
        if (itfTypes[i] == null) {
          if (i == 0) {
            throw new ParserException(
              binding, "No such client interface: " + s, null);
          } else {
            throw new ParserException(
              binding, "No such server interface: " + s, null);
          }
        }
      }
      s = binding.getAttribute("server", null);
    }
    // check subtyping relation between the two interfaces
    if (itfTypes[0] == null || itfTypes[1] == null) {
      return;
    }
    Class cc;
    try {
      cc = Class.forName(itfTypes[0].getAttribute("signature", null));
    } catch (ClassNotFoundException e) {
      throw new ParserException(binding, "Client Java interface not found", e);
    }
    Class sc;
    try {
      sc = Class.forName(itfTypes[1].getAttribute("signature", null));
    } catch (ClassNotFoundException e) {
      throw new ParserException(binding, "Server Java interface not found", e);
    }
    if (!cc.isAssignableFrom(sc)) {
      throw new ParserException(binding, "Incompatible interface types", null);
    }
  }

  /**
   * Checks a 'sharing' element.
   *
   * @param name the name of the current composite template.
   * @param sharing the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkSharing (final String name, final XMLElement sharing)
    throws ParserException
  {
    for (int i = 0; i < sharing.getSubElementCount(); ++i) {
      XMLElement elem = sharing.getSubElement(i);
      checkSharedComponent(name, elem);
    }
  }

  /**
   * Checks a 'shared-component' element.
   *
   * @param name the name of the current composite template.
   * @param shared the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkSharedComponent (final String name, final XMLElement shared)
    throws ParserException
  {
    try {
      findComponent(name, shared.getAttribute("path", null));
    } catch (ParserException e) {
      throw new ParserException(shared, "No such 'path' component", null);
    }
    try {
      findComponent(name, shared.getAttribute("ref", null));
    } catch (ParserException e) {
      throw new ParserException(shared, "No such 'ref' component", null);
    }
  }

  // external templates ------------------------------------------------------

  /**
   * Checks a 'primitive-template' element.
   *
   * @param tmpl the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkExternalTemplate (final XMLElement tmpl)
    throws ParserException
  {
    loadTypeDescriptor(tmpl.getAttribute("implements", null));
  }

  // controller definition ---------------------------------------------------

  /**
   * Checks a 'controller' element.
   *
   * @param cntrl the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkController (final XMLElement cntrl)
    throws ParserException
  {
    for (int i = 0; i < cntrl.getSubElementCount(); ++i) {
      XMLElement elem = cntrl.getSubElement(i);
      if (elem.tag.equals("attributes")) {
        checkAttributes(elem);
      }
    }
  }

  // attributes --------------------------------------------------------------

  /**
   * Checks an 'attributes' element.
   *
   * @param attrs the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkAttributes (final XMLElement attrs)
    throws ParserException
  {
    // checks that the 'signature' Java interface exists
    String s = attrs.getAttribute("signature", null);
    Class c;
    try {
      c = Class.forName(s);
    } catch (ClassNotFoundException e) {
      throw new ParserException(attrs, "Class not found: "+ s, e);
    }
    // checks the attributes
    for (int i = 0; i < attrs.getSubElementCount(); ++i) {
      XMLElement elem = attrs.getSubElement(i);
      checkAttribute(c, elem);
    }
  }

  /**
   * Checks an 'attribute' element.
   *
   * @param c the attribute controller class.
   * @param attr the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkAttribute (final Class c, final XMLElement attr)
    throws ParserException
  {
    // checks that the 'name' attribute exists,
    // and checks that the type of 'value' is correct.
    String attrName = attr.getAttribute("name", null);
    String attrValue = attr.getAttribute("value", null);
    try {
      String getterName = "get" + attrName;
      Method getter = c.getMethod(getterName, new Class[0]);
      Class attrType = getter.getReturnType();
      if (attrType.isPrimitive()) {
        if (attrType.equals(Integer.TYPE)) {
          Integer.valueOf(attrValue);
        } else if (attrType.equals(Long.TYPE)) {
          Long.valueOf(attrValue);
        } else if (attrType.equals(Float.TYPE)) {
          Float.valueOf(attrValue);
        } else if (attrType.equals(Double.TYPE)) {
          Double.valueOf(attrValue);
        } else if (attrType.equals(Byte.TYPE)) {
          Byte.valueOf(attrValue);
        } else if (attrType.equals(Character.TYPE)) {
          if (attrValue.length() != 1) {
            throw new Exception("Bad char value: " + attrValue);
          }
        } else if (attrType.equals(Short.TYPE)) {
          Short.valueOf(attrValue);
        } else if (attrType.equals(Boolean.TYPE)) {
          if (!attrValue.equals("true") && !attrValue.equals("false")) {
            throw new Exception("Bad boolean value: " + attrValue);
          }
        } else {
          throw new Exception("Unexpected case");
        }
      } else if (attrType != String.class) {
        throw new Exception("Unsupported attribute type: " + attrType);
      }
    } catch (Exception e) {
      throw new ParserException(
        attr, "No such attribute or bad value type '" + attrName + "'", e);
    }
  }

  // type checking methods ---------------------------------------------------

  /**
   * Checks that a component type is a sub type of another one.
   *
   * @param provided a component type name.
   * @param required a component type name.
   * @throws ParserException if the <tt>provided</tt> type is a not a sub type
   *      of the <tt>required</tt> type, or if the two types cannot be compared.
   */

  private void checkSubType (final String provided, final String required)
    throws ParserException
  {
    XMLElement p = loadTypeDescriptor(provided);
    XMLElement r = loadTypeDescriptor(required);
    XMLElement pp = p.getSubElement("provides");
    XMLElement pr = p.getSubElement("requires");
    XMLElement rp = r.getSubElement("provides");
    XMLElement rr = r.getSubElement("requires");
    checkSubType(p, pp, r, rp);
    checkSubType(r, rr, p, pr);
  }

  /**
   * Checks that a set of interface types is a sub type of another one.
   *
   * @param pType a &lt;component-type&gt; element.
   * @param pItfs the &lt;provides&gt; or &lt;requires&gt; sub element of
   *      <tt>pType</tt>.
   * @param rType a &lt;component-type&gt; element.
   * @param rItfs the &lt;requires&gt; or &lt;provides&gt; sub element of
   *      <tt>rType</tt>.
   * @throws ParserException if the <tt>pItfs</tt> type set a not a sub type
   *      of the <tt>rItf</tt> type set, or if the two sets cannot be compared.
   */

  private void checkSubType (
    final XMLElement pType,
    final XMLElement pItfs,
    final XMLElement rType,
    final XMLElement rItfs) throws ParserException
  {
    if (rItfs == null) {
      return;
    }
    // checks that, for each interface type i in rItfs, there is an interface
    // type j in pItfs with the same name, and whose Java type is a sub type of
    // the Java type of i.
    for (int i = 0; i < rItfs.getSubElementCount(); ++i) {
      XMLElement ritf = rItfs.getSubElement(i);
      String rname = ritf.getAttribute("name", null);
      String rsig = ritf.getAttribute("signature", null);
      Class rc;
      try {
        rc = Class.forName(rsig);
      } catch (ClassNotFoundException e) {
        throw new ParserException(
          rType, new ParserException(ritf, "Class not found: " + rsig, e));
      }
      boolean ok = false;
      if (pItfs != null) {
        for (int j = 0; j < pItfs.getSubElementCount(); ++j) {
          XMLElement pitf = pItfs.getSubElement(j);
          String pname = pitf.getAttribute("name", null);
          if (pname.equals(rname)) {
            String psig = pitf.getAttribute("signature", null);
            Class pc;
            try {
              pc = Class.forName(psig);
            } catch (ClassNotFoundException e) {
              throw new ParserException(
                pType, new ParserException(pitf, "Class not found: " + psig, e));
            }
            if (!rc.isAssignableFrom(pc)) {
              throw new ParserException(
                pc + " is not a sub type of " + rc, null);
            }
            ok = true;
            break;
          }
        }
      }
      if (!ok) {
        throw new ParserException("'" + rname + "' interface missing", null);
      }
    }
  }

  /**
   * Checks if the specified component contains a sub component at the given
   * path.
   *
   * @param template a template description name.
   * @param path a path designating a direct or indirect sub component in
   *      <tt>template</tt>.
   * @throws ParserException if the specified component does not contain a sub
   *      component at the given path.
   */

  private void findComponent (final String template, final String path)
    throws ParserException
  {
    String head;
    String tail;
    int index = path.indexOf('/');
    if (index == -1) {
      head = path;
      tail = null;
    } else {
      head = path.substring(0, index);
      tail = path.substring(index + 1);
    }
    XMLElement tmpl = delegate.loadTemplateDescriptor(template);
    XMLElement cont = tmpl.getSubElement("composite-content");
    XMLElement comps = cont.getSubElement("components");
    XMLElement comp = findComponent(comps, head);
    if (tail != null) {
      findComponent(comp.getAttribute("implementation", null), tail);
    }
  }

  /**
   * Returns the &lt;component&gt; element describing the specified sub
   * component.
   *
   * @param comps a &lt;components&gt; element.
   * @param name a sub component name.
   * @return the &lt;component&gt; element describing the specified sub
   *      component.
   * @throws ParserException if the sub component description cannot be found.
   */

  private XMLElement findComponent (final XMLElement comps, final String name)
    throws ParserException
  {
    for (int i = 0; i < comps.getSubElementCount(); ++i) {
      XMLElement comp = comps.getSubElement(i);
      if (comp.getAttribute("name", null).equals(name)) {
        return comp;
      }
    }
    return null;
  }

  /**
   * Returns the &lt;interface-type&gt; element describing the specified
   * interface.
   *
   * @param type a component type name.
   * @param name an interface name.
   * @param provided <tt>true</tt> to look for a provided interface,
   *      <tt>false</tt> to look for a required interface.
   * @return the &lt;interface-type&gt; element describing the specified
   *      interface.
   *  @throws ParserException if the interface description cannot be found.
   */

  private XMLElement findInterface (
    final String type,
    final String name,
    final boolean provided) throws ParserException
  {
    XMLElement t = loadTypeDescriptor(type);
    XMLElement itfs = t.getSubElement(provided ? "provides" : "requires");
    if (itfs != null) {
      for (int i = 0; i < itfs.getSubElementCount(); ++i) {
        XMLElement itf = itfs.getSubElement(i);
        if (itf.getAttribute("cardinality", "single").equals("collection")) {
          if (name.startsWith(itf.getAttribute("name", null))) {
            return itf;
          }
        } else {
          if (itf.getAttribute("name", null).equals(name)) {
            return itf;
          }
        }
      }
    }
    return null;
  }
}
