/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.objectweb.fractal.api.control.BindingController;

import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.Logger;

/**
 * A {@link Loader Loader} that delegates parts of its work to another {@link
 * Loader Loader}.
 */

public abstract class DelegatingLoader
  implements Loader, BindingController
{

  /**
   * The {@link Loader Loader} to which this loader can delegate work.
   */

  protected Loader delegate;

  /**
   * The optional logger factory used to get a logger for this component.
   */

  protected LoggerFactory loggerFactory;

  /**
   * The logger used to log messages. May be <tt>null</tt>.
   */

  protected Logger logger;

  // -------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // -------------------------------------------------------------------------

  public String[] listFc () {
    return new String[] { "delegate", "logger-factory" };
  }

  public Object lookupFc (final String clientItfName) {
    if (clientItfName.equals("delegate")) {
      return delegate;
    } else if (clientItfName.equals("logger-factory")) {
      return loggerFactory;
    }
    return null;
  }

  public void bindFc (final String clientItfName, final Object serverItf) {
    if (clientItfName.equals("delegate")) {
      delegate = (Loader)serverItf;
    } else if (clientItfName.equals("logger-factory")) {
      loggerFactory = (LoggerFactory)serverItf;
      logger = loggerFactory.getLogger(getClass().getName());
    }
  }

  public void unbindFc (final String clientItfName) {
    if (clientItfName.equals("delegate")) {
      delegate = null;
    } else if (clientItfName.equals("logger-factory")) {
      loggerFactory = null;
      logger = null;
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the Loader interface
  // -------------------------------------------------------------------------

  public XMLElement loadTypeDescriptor (final String name)
    throws ParserException
  {
    XMLElement desc = delegate.loadTypeDescriptor(name);
    try {
      desc = handleTypeDescriptor(name, desc);
    } catch (ParserException e) {
      throw new ParserException(desc, e);
    }
    return desc;
  }

  public XMLElement loadTemplateDescriptor (final String name)
    throws ParserException
  {
    XMLElement desc = delegate.loadTemplateDescriptor(name);
    try {
      desc = handleTemplateDescriptor(name, desc);
    } catch (ParserException e) {
      throw new ParserException(desc, e);
    }
    return desc;
  }

  // -------------------------------------------------------------------------
  // Utility methods
  // -------------------------------------------------------------------------

  /**
   * Handles a type descriptor that has just been loaded by the {@link #delegate
   * delegate}.
   *
   * @param name the name of the type descriptor.
   * @param desc a type descriptor.
   * @return the given type descriptor, possibly modified by this method.
   * @throws ParserException if the type descriptor is invalid.
   */

  protected abstract XMLElement handleTypeDescriptor (
    final String name,
    final XMLElement desc) throws ParserException;

  /**
   * Handles a template descriptor that has just been loaded by the {@link
   * #delegate delegate}.
   *
   * @param name the name of the template descriptor.
   * @param desc a template descriptor.
   * @return the given template descriptor, possibly modified by this method.
   * @throws ParserException if the template descriptor is invalid.
   */

  protected abstract XMLElement handleTemplateDescriptor (
    final String name,
    final XMLElement desc) throws ParserException;
}
