/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.taskdefs.MatchingTask;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * An Ant task to check some type and template definitions, and to store them
 * in an output directory.
 */

public class FractalcTask extends MatchingTask {

  /**
   * The directory that contains the definitions to be processed.
   */

  private File src;

  /**
   * The directory where the processed definitions must be stored.
   */

  private File dst;

  /**
   * If messages must be logged.
   */

  private boolean verbose;

  /**
   * If the definitions must be checked before being copied to <tt>dst</tt>.
   */

  private boolean verify;

  /**
   * If the equivalent kilim definitions must be generated and stored in
   * <tt>dst</tt>.
   */

  private boolean kilim;

  /**
   * Sets the directory that contains the definitions to be processed.
   *
   * @param src the directory that contains the definitions to be processed.
   */

  public void setSrcdir (final File src) {
    this.src = src;
  }

  /**
   * Sets the directory where the processed definitions must be stored.
   *
   * @param dst the directory where the processed definitions must be stored.
   */

  public void setDestdir (final File dst) {
    this.dst = dst;
  }

  /**
   * Sets the verbose option.
   *
   * @param verbose the verbose option. If this option is <tt>true</tt> messages
   *      are printed for each processed file.
   */

  public void setVerbose (final boolean verbose) {
    this.verbose = verbose;
  }

  /**
   * Sets the verify option.
   *
   * @param verify the verify option. If this option is <tt>true</tt> the type
   *      and template definitions are checked after they have been copied to
   *      the destination directory.
   */

  public void setVerify (final boolean verify) {
    this.verify = verify;
  }

  /**
   * Sets the kilim option.
   *
   * @param kilim the kilim option. If this option is <tt>true</tt> kilim
   *      definitions equivalent to the processed definitions are also generated
   *      in the output directory.
   */

  public void setKilim (final boolean kilim) {
    this.kilim = kilim;
  }

  /**
   * Executes this task.
   */

  public void execute () {
    if (src == null) {
      throw new BuildException("srcdir must be specified");
    }
    if (dst == null) {
      throw new BuildException("destdir must be specified");
    }

    Map types = new HashMap();
    Map templates = new HashMap();

    DirectoryScanner ds = getDirectoryScanner(src);
    String[] files = ds.getIncludedFiles();
    for (int i = 0; i < files.length; ++i) {
      processFile(files[i], types, templates);
    }

    ArrayList errors = new ArrayList();

    if (verify) {
      BasicLoader bl = new BasicLoader();
      VerifyingLoader vl = new VerifyingLoader();
      vl.bindFc("delegate", bl);
      ResolvingLoader rl = new ResolvingLoader();
      rl.bindFc("delegate", vl);
      CheckingLoader cl = new CheckingLoader();
      cl.bindFc("delegate", rl);
      Loader loader = cl;

      Iterator i = types.keySet().iterator();
      for (int j = 0; j < 2; ++j) {
        while (i.hasNext()) {
          String name = (String)i.next();
          String s = name.replace('.', File.separatorChar) + ".fractal";
          try {
            if (verbose) {
              log("verifying " + name);
            }
            if (j == 0) {
              loader.loadTypeDescriptor(name);
            } else {
              loader.loadTemplateDescriptor(name);
            }
          } catch (Exception e) {
            log("ERROR WHILE VERIFYING " + name + ":");
            e.printStackTrace();
            log("");
            errors.add(new File(dst, s));
          }
        }
        i = templates.keySet().iterator();
      }
    }

    if (kilim) {
      BasicLoader bl = new BasicLoader();
      ResolvingLoader rl = new ResolvingLoader(true);
      rl.bindFc("delegate", bl);
      KilimLoader kl = new KilimLoader();
      kl.bindFc("delegate", rl);
      Loader loader = kl;

      Iterator i = types.keySet().iterator();
      for (int j = 0; j < 2; ++j) {
        while (i.hasNext()) {
          String name = (String)i.next();
          String s = name.replace('.', File.separatorChar) + ".fractal";
          try {
            if (verbose) {
              log("converting " + name);
            }
            if (j == 0) {
              save(loader.loadTypeDescriptor(name), ".kilim");
            } else {
              save(loader.loadTemplateDescriptor(name), ".kilim");
            }
          } catch (Exception e) {
            log("ERROR WHILE CONVERTING " + name + ":");
            e.printStackTrace();
            log("");
            errors.add(new File(dst, s));
          }
        }
        i = templates.keySet().iterator();
      }
    }

    if (errors.size() > 0) {
      for (int i = 0; i < errors.size(); ++i) {
        ((File)errors.get(i)).delete();
      }
      throw new BuildException();
    }
  }

  /**
   * Process an input file. The definitions contained in the specified file
   * are copied to the destination directory, one per file. They are also stored
   * in the given maps.
   *
   * @param file the file to be processed.
   * @param types a map associating type names to type descriptors.
   * @param templates a map associating template names to template descriptors.
   */

  private void processFile (
    final String file,
    final Map types,
    final Map templates)
  {
    File f = new File(src, file);
    LineNumberReader r;
    try {
      r = new LineNumberReader(new FileReader(f));
    } catch (IOException ioe) {
      log("Cannot read " + f + ":\n" + ioe.getMessage());
      throw new BuildException();
    }
    while (true) {
      XMLElement e;
      try {
        e = XMLElement.parse(r, false);
      } catch (IOException ioe) {
        log("File " + f + ":\n" + ioe.getMessage());
        throw new BuildException();
      } catch (ParserException pe) {
        log("File " + f + ", line " + pe.getLine() + ":\n" + pe.getMessage());
        throw new BuildException();
      }
      if (e == null) {
        return;
      } else {
        e.setAttribute("[source]", file + ":" + Integer.toString(e.line));
      }
      if (!e.tag.equals("component-type") &&
          !e.tag.equals("primitive-template") &&
          !e.tag.equals("composite-template") &&
          !e.tag.equals("external-template"))
      {
        log("WARNING: unrecognized tag '" + e.tag+ "'");
        continue;
      }
      String name;
      try {
        name = e.getAttribute("name", null);
      } catch (ParserException pe) {
        log("File " + f + ":\n" + pe.getMessage());
        throw new BuildException();
      }
      String s = name.replace('.', File.separatorChar) + ".fractal";
      File g = new File(dst, s);
      if (!g.exists() || g.lastModified() < f.lastModified()) {
        if (verbose) {
          log("generating " + dst + File.separatorChar + s);
        }
        save(e, ".fractal");
        if (e.tag.equals("component-type")) {
          types.put(name, e);
        } else {
          templates.put(name, e);
        }
      }
    }
  }

  /**
   * Saves the given XML element to the destination directory.
   *
   * @param elem an XML element.
   * @param ext the file extension to be used to save this XML element.
   */

  private void save (final XMLElement elem, final String ext) {
    String name;
    try {
      name = elem.getAttribute("name", null);
    } catch (ParserException e) {
      log("Unexpected error");
      throw new BuildException();
    }
    File f = new File(dst, name.replace('.', File.separatorChar) + ext);
    File p = f.getParentFile();
    if (!p.exists()) {
      p.mkdirs();
    }
    try {
      FileWriter fw = new FileWriter(f);
      fw.write("<?xml version=\"1.0\" encoding=\"ISO-8859-1\" ?>\n");
      if (ext.equals(".fractal")) {
        fw.write("<!DOCTYPE ");
        fw.write(elem.tag);
        fw.write(" PUBLIC \"-//objectweb.org//DTD Fractal ADL 1.0//EN\" ");
        fw.write("\"classpath://fractal-adl.dtd\">\n\n");
        try {
          String source = elem.getAttribute("[source]", null);
          elem.setAttribute("[source]", null);
          fw.write("<!-- copied from ");
          fw.write(source);
          fw.write(" -->\n\n");
        } catch (ParserException ignored) {
        }
      } else {
        fw.write("<!DOCTYPE template PUBLIC ");
        fw.write("\"-//objectweb.org//TEXT template description 4.0//EN\" ");
        fw.write("\"classpath://kilim_template.dtd\">\n\n");
      }
      elem.write(fw);
      fw.close();
    } catch (IOException ioe) {
      log("Cannot write " + f + ":\n" + ioe.getMessage());
      throw new BuildException();
    }
  }
}
