/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import java.lang.reflect.Method;

/**
 * A {@link DelegatingLoader DelegatingLoader} that converts the definitions
 * loaded by its delegate loader to Kilim 2 templates.
 */

public class KilimLoader extends DelegatingLoader {

  // -------------------------------------------------------------------------
  // Overriden methods
  // -------------------------------------------------------------------------

  protected XMLElement handleTypeDescriptor (
    final String name,
    final XMLElement desc) throws ParserException
  {
    try {
      return convertComponentType(desc);
    } catch (ParserException e) {
      throw new ParserException(desc, e);
    }
  }

  protected XMLElement handleTemplateDescriptor (
    final String name,
    final XMLElement desc) throws ParserException
  {
    try {
      if (desc.tag.equals("primitive-template")) {
        try {
          return convertPrimitiveTemplate(true, desc);
        } catch (ConversionException e) {
          System.err.println("Warning: cannot preserve inheritance");
          return convertPrimitiveTemplate(false, desc);
        }
      } else if (desc.tag.equals("composite-template")) {
        try {
          return convertCompositeTemplate(true, desc);
        } catch (ConversionException e) {
          System.err.println("Warning: cannot preserve inheritance");
          return convertCompositeTemplate(false, desc);
        }
      } else if (desc.tag.equals("external-template")) {
        throw new ParserException("TODO not yet implemented", null);
      } else {
        return null;
      }
    } catch (ParserException e) {
      throw new ParserException(desc, e);
    }
  }

  // -------------------------------------------------------------------------
  // Utility methods
  // -------------------------------------------------------------------------

  // component types ---------------------------------------------------------

  /**
   * Converts a 'component-type' element.
   *
   * @param type the element to be converted.
   * @return the converted element.
   * @throws ParserException if the given element cannot be converted.
   */

  private XMLElement convertComponentType (final XMLElement type)
    throws ParserException
  {
    XMLElement r = new XMLElement("template");
    r.setAttribute("name", type.getAttribute("name", null));
    String extend = type.getAttribute("extends", "");
    if (extend.length() > 0) {
      r.setAttribute("extends", extend);
    }
    for (int i = 0; i < type.getSubElementCount(); ++i) {
      XMLElement elem = type.getSubElement(i);
      convertInterfaceTypes(true, elem, r);
    }
    return r;
  }

  /**
   * Converts a 'provides' or 'requires' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param types the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertInterfaceTypes (
    final boolean inheritance,
    final XMLElement types,
    final XMLElement r) throws ParserException
  {
    boolean provided = types.tag.equals("provides");
    for (int i = 0; i < types.getSubElementCount(); ++i) {
      XMLElement elem = types.getSubElement(i);
      if (!inheritance || !elem.getAttribute("[src]", "this").equals("super")) {
        convertInterfaceType(provided, elem, r);
      }
    }
  }

  /**
   * Converts an 'interface-type' element.
   *
   * @param provided if the given element corresponds to a provided interface.
   * @param type the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertInterfaceType (
    final boolean provided,
    final XMLElement type,
    final XMLElement r) throws ParserException
  {
    XMLElement p = new XMLElement("port");
    p.setAttribute("name", type.getAttribute("name", null));
    p.setAttribute("role", provided ? "offered" : "required");
    boolean any = type.getAttribute("cardinality", "single").equals("collection");
    boolean opt = type.getAttribute("contingency", "mandatory").equals("optional");
    if (any) {
      p.setAttribute("arity", "any");
    } else if (opt) {
      p.setAttribute("arity", "optional");
    } else {
      p.setAttribute("arity", "1");
    }
    XMLElement t = new XMLElement("type");
    t.addSubElement(
      new XMLElement("<![CDATA[" + type.getAttribute("signature", null) + "]]>"));
    p.addSubElement(t);
    r.addSubElement(p);
  }

  // templates ---------------------------------------------------------------

  /**
   * Converts a 'primitive-template' or 'composite-template' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param tmpl the element to be converted.
   * @return the converted element.
   * @throws ParserException if the given element cannot be converted.
   */

  private XMLElement convertTemplate (
    final boolean inheritance,
    final XMLElement tmpl) throws ParserException
  {
    XMLElement r = new XMLElement("template");
    r.setAttribute("name", tmpl.getAttribute("name", null));
    String implement = tmpl.getAttribute("implements", "");
    String extend = tmpl.getAttribute("extends", "");
    if (inheritance && extend.length() > 0) {
      r.setAttribute("extends", extend);
      if (tmpl.getAttribute("[implements]", "").length() > 0) {
        // cannot preserve inheritance relations if
        // both 'extends' and 'implements' are defined
        throw new ConversionException();
      }
    } else {
      r.setAttribute("extends", implement);
    }
    return r;
  }

  // primitive templates -----------------------------------------------------

  /**
   * Converts a 'primitive-template' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param tmpl the element to be converted.
   * @return the converted element.
   * @throws ParserException if the given element cannot be converted.
   */

  private XMLElement convertPrimitiveTemplate (
    final boolean inheritance,
    final XMLElement tmpl) throws ParserException
  {
    XMLElement r = convertTemplate(inheritance, tmpl);
    for (int i = 0; i < tmpl.getSubElementCount(); ++i) {
      XMLElement elem = tmpl.getSubElement(i);
      if (elem.tag.equals("primitive-content")) {
        convertPrimitiveContent(inheritance, elem, r);
      } else {
        convertController(inheritance, elem, r);
      }
    }
    addBindsAndTriggers(inheritance, tmpl, r);
    return r;
  }

  /**
   * Converts a 'primitive-content' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param content the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertPrimitiveContent (
    final boolean inheritance,
    final XMLElement content,
    final XMLElement r) throws ParserException
  {
    String src = content.getAttribute("[src]", "this");
    if (!inheritance || src.equals("this")) {
      XMLElement p = new XMLElement("port");
      p.setAttribute("name", "content");
      p.setAttribute("role", "offered");
      p.setAttribute("arity", "1");
      r.addSubElement(p);

      XMLElement ref = new XMLElement("reference");
      ref.setAttribute("target", "content-provider");
      XMLElement b = new XMLElement("bind");
      b.setAttribute("port", "content");
      b.addSubElement(ref);
      r.addSubElement(b);
    }
    if (!inheritance || !src.equals("super")) {
      XMLElement c = new XMLElement("class");
      c.setAttribute("name", content.getAttribute("class", null));
      XMLElement n = new XMLElement("new");
      n.addSubElement(c);
      XMLElement q = new XMLElement("provider");
      q.setAttribute("name", "content-provider");
      q.addSubElement(n);
      r.addSubElement(q);
    }
  }

  /**
   * Creates the bind and trigger elements corresponding to the given template.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param tmpl a 'primitive-template' element.
   * @param r the element to which the bind and trigger elements must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void addBindsAndTriggers (
    final boolean inheritance,
    final XMLElement tmpl,
    final XMLElement r) throws ParserException
  {
    String extend = tmpl.getAttribute("extends", "");
    String implement = tmpl.getAttribute("implements", "");
    if (inheritance && extend.length() > 0) {
      String superImplement = tmpl.getAttribute("[implements]", "");
      if (superImplement.length() == 0 || implement.equals(superImplement)) {
        // bindings and triggers are defined in the super Kilim template,
        // so they must not be redefined in this template
        return;
      } else {
        // in this case we must just define in this template the bindings
        // and triggers that are not already defined in the super Kilim
        // template (this can be done by comparing the two component types)
        throw new ParserException("TODO not yet implemented", null);
      }
    }
    XMLElement type = delegate.loadTypeDescriptor(implement);
    for (int i = 0; i < type.getSubElementCount(); ++i) {
      XMLElement elem = type.getSubElement(i);
      if (elem.tag.equals("provides")) {
        addBinds(elem, r);
      } else {
        addTriggers(elem, r);
      }
    }
  }

  /**
   * Creates the bind elements corresponding to the given interface types.
   *
   * @param types a 'provides' element.
   * @param r the element to which the bind elements must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void addBinds (final XMLElement types, final XMLElement r)
    throws ParserException
  {
    for (int i = 0; i < types.getSubElementCount(); ++i) {
      XMLElement elem = types.getSubElement(i);
      addBind(elem, r);
    }
  }

  /**
   * Creates the bind element corresponding to the given interface type.
   *
   * @param type an 'interface-type' element.
   * @param r the element to which the bind element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void addBind (final XMLElement type, final XMLElement r)
    throws ParserException
  {
    XMLElement bind = new XMLElement("bind");
    bind.setAttribute("port", type.getAttribute("name", null));
    XMLElement ref = new XMLElement("reference");
    ref.setAttribute("target", "content");
    bind.addSubElement(ref);
    r.addSubElement(bind);
  }

  /**
   * Creates the trigger elements corresponding to the given interface types.
   *
   * @param types a 'requires' element.
   * @param r the element to which the trigger elements must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void addTriggers (final XMLElement types, final XMLElement r)
    throws ParserException
  {
    for (int i = 0; i < types.getSubElementCount(); ++i) {
      XMLElement elem = types.getSubElement(i);
      addTrigger(elem, r);
    }
  }

  /**
   * Creates the trigger element corresponding to the given interface type.
   *
   * @param type an 'interface-type' element.
   * @param r the element to which the trigger element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void addTrigger (final XMLElement type, final XMLElement r)
    throws ParserException
  {
    String name = type.getAttribute("name", null);
    XMLElement trigger = new XMLElement("trigger");
    trigger.setAttribute("source", "content");
    trigger.setAttribute("event", "bind");
    XMLElement call = new XMLElement("call");
    call.setAttribute("method", "addFcBinding");
    call.addSubElement(new XMLElement("event-source"));
    XMLElement param = new XMLElement("param");
    param.setAttribute("type", "java.lang.String");
    XMLElement prop = new XMLElement("property");
    prop.setAttribute("type", "string");
    prop.addSubElement(new XMLElement("<![CDATA[" + name + "]]>"));
    param.addSubElement(prop);
    call.addSubElement(param);
    param = new XMLElement("param");
    param.setAttribute("type", "java.lang.Object");
    XMLElement ref = new XMLElement("reference");
    ref.setAttribute("target", name);
    param.addSubElement(ref);
    call.addSubElement(param);
    trigger.addSubElement(call);
    r.addSubElement(trigger);
  }

  // composite templates -----------------------------------------------------

  /**
   * Converts a 'composite-template' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param tmpl the element to be converted.
   * @return the converted element.
   * @throws ParserException if the given element cannot be converted.
   */

  private XMLElement convertCompositeTemplate (
    final boolean inheritance,
    final XMLElement tmpl) throws ParserException
  {
    XMLElement r = convertTemplate(inheritance, tmpl);
    for (int i = 0; i < tmpl.getSubElementCount(); ++i) {
      XMLElement elem = tmpl.getSubElement(i);
      if (elem.tag.equals("composite-content")) {
        convertCompositeContent(inheritance, elem, r);
      } else {
        convertController(inheritance, elem, r);
      }
    }
    return r;
  }

  /**
   * Converts a 'composite-content' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param content the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertCompositeContent (
    final boolean inheritance,
    final XMLElement content,
    final XMLElement r) throws ParserException
  {
    XMLElement components = null;
    for (int i = 0; i < content.getSubElementCount(); ++i) {
      XMLElement elem = content.getSubElement(i);
      if (elem.tag.equals("components")) {
        components = elem;
        convertComponents(inheritance, elem, r);
      } else if (elem.tag.equals("bindings")) {
        convertBindings(inheritance, components, elem, r);
      } else if (elem.tag.equals("sharing")) {
        convertSharing(elem, r);
      }
    }
  }

  /**
   * Converts a 'components' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param comps the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertComponents (
    final boolean inheritance,
    final XMLElement comps,
    final XMLElement r) throws ParserException
  {
    for (int i = 0; i < comps.getSubElementCount(); ++i) {
      XMLElement elem = comps.getSubElement(i);
      convertComponent(inheritance, elem, r);
    }
  }

  /**
   * Converts a 'component' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param comp the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertComponent (
    final boolean inheritance,
    final XMLElement comp,
    final XMLElement r) throws ParserException
  {
    String src = comp.getAttribute("[src]", "this");
    if (inheritance) {
      if (src.equals("super")) {
        return;
      }
      if (src.equals("merge")) {
        if (comp.getAttribute("[type]", "").length() > 0) {
          // Kilim <slot> elements cannot be overriden in sub templates, so we
          // cannot convert this composite template while preserving the Fractal
          // inheritance relations
          throw new ConversionException();
        }
      }
    }
    String n = comp.getAttribute("name", null);
    String m = "_" + n;

    if (!inheritance || src.equals("this")) {
      XMLElement s = new XMLElement("slot");
      s.setAttribute("name", comp.getAttribute("name", null));
      String typeName = comp.getAttribute("type", "");

      XMLElement type = delegate.loadTypeDescriptor(typeName);
      for (int i = 0; i < type.getSubElementCount(); ++i) {
        XMLElement elem = type.getSubElement(i);
        convertInterfaceTypes(false, elem, s);
      }
      r.addSubElement(s);
    }

    String impl = comp.getAttribute("implementation", "");
    if (impl.length() > 0) {
      XMLElement i = new XMLElement("instance");
      i.setAttribute("name", m);
      i.setAttribute("template", impl);
      r.addSubElement(i);

      if (!inheritance || comp.getAttribute("[implementation]", "").length() == 0) {
        XMLElement p = new XMLElement("plug");
        p.setAttribute("slot", n);
        p.setAttribute("instance", m);
        r.addSubElement(p);
      }
    }
  }

  /**
   * Converts a 'bindings' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param components a 'components' element, to find the bound components.
   * @param bindings the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertBindings (
    final boolean inheritance,
    final XMLElement components,
    final XMLElement bindings,
    final XMLElement r) throws ParserException
  {
    for (int i = 0; i < bindings.getSubElementCount(); ++i) {
      XMLElement elem = bindings.getSubElement(i);
      convertBinding(inheritance, components, elem, r);
    }
  }

  /**
   * Converts a 'binding' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param components a 'components' element, to find the bound components.
   * @param binding the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertBinding (
    final boolean inheritance,
    final XMLElement components,
    final XMLElement binding,
    final XMLElement r) throws ParserException
  {
    String c = binding.getAttribute("client", null).replace('.', '/');
    String s = binding.getAttribute("server", null).replace('.', '/');
    String cc = c.substring(0, c.indexOf('/'));
    String sc = s.substring(0, s.indexOf('/'));
    XMLElement xcc = null;
    XMLElement xsc = null;
    // if one of the component is not defined in this template (it will be
    // defined in sub templates), do not convert this binding
    if (!cc.equals("this")) {
      xcc = getComponent(components, cc);
      if (xcc == null || xcc.getAttribute("implementation", "").length() == 0) {
        return;
      }
    }
    if (!sc.equals("this")) {
      xsc = getComponent(components, sc);
      if (xsc == null || xsc.getAttribute("implementation", "").length() == 0) {
        return;
      }
    }
    if (inheritance) {
      String src = binding.getAttribute("[src]", "this");
      if (src.equals("super")) {
        // tests if this binding was converted in the super template or not
        boolean converted = true;
        if (!cc.equals("this")) {
          if (xcc.getAttribute("[implementation]", "").length() == 0) {
            converted = false;
          }
        }
        if (!sc.equals("this")) {
          if (xsc.getAttribute("[implementation]", "").length() == 0) {
            converted = false;
          }
        }
        if (converted) {
          // this binding has already been converted in the super template:
          // do not convert it again
          return;
        }
      } else if (src.equals("merge")) {
        // Kilim <bind> elements cannot be overriden in sub templates, so we
        // cannot convert this composite template while preserving the Fractal
        // inheritance relations
        throw new ConversionException();
      }
    }
    XMLElement b = new XMLElement("bind");
    if (c.startsWith("this/")) {
      c = c.substring(5);
    }
    b.setAttribute("port", c);
    XMLElement ref = new XMLElement("reference");
    if (s.startsWith("this/")) {
      s = s.substring(5);
    }
    ref.setAttribute("target", s);
    b.addSubElement(ref);
    r.addSubElement(b);
  }

  /**
   * Returns the 'component' element of the given name.
   *
   * @param components a 'components' element.
   * @param name the name of a sub component.
   * @return the 'component' sub element of <tt>components</tt> whose name is
   *      equal to the given name, or <tt>null</tt> if there is no such
   *      component.
   * @throws ParserException if a problem occurs.
   */

  private XMLElement getComponent (
    final XMLElement components,
    final String name) throws ParserException
  {
    for (int i = 0; i < components.getSubElementCount(); ++i) {
      XMLElement elem = components.getSubElement(i);
      if (elem.getAttribute("name", null).equals(name)) {
        return elem;
      }
    }
    return null;
  }

  /**
   * Converts a 'sharing' element.
   *
   * @param sharing the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertSharing (final XMLElement sharing, final XMLElement r)
    throws ParserException
  {
    for (int i = 0; i < sharing.getSubElementCount(); ++i) {
      XMLElement elem = sharing.getSubElement(i);
      convertSharedComponent(elem, r);
    }
  }

  /**
   * Converts a 'shared-component' element.
   *
   * @param shared the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertSharedComponent (
    final XMLElement shared,
    final XMLElement r) throws ParserException
  {
    // TODO how shared components can be converted to Kilim?
  }

  // controller definition ---------------------------------------------------

  /**
   * Converts a 'controller' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param cntrl the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertController (
    final boolean inheritance,
    final XMLElement cntrl,
    final XMLElement r) throws ParserException
  {
    for (int i = 0; i < cntrl.getSubElementCount(); ++i) {
      XMLElement elem = cntrl.getSubElement(i);
      if (elem.tag.equals("attributes")) {
        convertAttributes(inheritance, elem, r);
      }
    }
  }

  // attributes --------------------------------------------------------------

  /**
   * Converts an 'attributes' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param attrs the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertAttributes (
    final boolean inheritance,
    final XMLElement attrs,
    final XMLElement r) throws ParserException
  {
    String name = attrs.getAttribute("signature", null);
    Class c;
    try {
      c = Class.forName(name);
    } catch (ClassNotFoundException e) {
      throw new ParserException(
        attrs, "Cannot find attribute controller class: " + name, e);
    }
    for (int i = 0; i < attrs.getSubElementCount(); ++i) {
      XMLElement elem = attrs.getSubElement(i);
      convertAttribute(inheritance, c, elem, r);
    }

    XMLElement trigger = new XMLElement("trigger");
    trigger.setAttribute("source", "content");
    trigger.setAttribute("event", "bind");
    for (int i = 0; i < attrs.getSubElementCount(); ++i) {
      XMLElement elem = attrs.getSubElement(i);
      if (inheritance && !elem.getAttribute("[src]", "this").equals("this")) {
        continue;
      }
      String attr = elem.getAttribute("name", null);

      XMLElement call = new XMLElement("call");
      call.setAttribute("method", "set" + attr);
      XMLElement ref = new XMLElement("event-source");
      call.addSubElement(ref);
      XMLElement param = new XMLElement("param");
      String typ = getAttributeType(c, elem, attr);
      param.setAttribute("type", typ.equals("string") ? "java.lang.String" : typ);
      ref = new XMLElement("reference");
      ref.setAttribute("target", attr);
      param.addSubElement(ref);
      call.addSubElement(param);

      trigger.addSubElement(call);
    }
    if (trigger.getSubElementCount() > 0) {
      r.addSubElement(trigger);
    }
  }

  /**
   * Converts an 'attribute' element.
   *
   * @param inheritance if the inheritance relations must be preserved.
   * @param attrControllerClass the attribute controller class.
   * @param attr the element to be converted.
   * @param r the element to which the converted element must be added.
   * @throws ParserException if the given element cannot be converted.
   */

  private void convertAttribute (
    final boolean inheritance,
    final Class attrControllerClass,
    final XMLElement attr,
    final XMLElement r) throws ParserException
  {
    if (inheritance && attr.getAttribute("[src]", "this").equals("super")) {
      return;
    }
    XMLElement p = new XMLElement("property");
    String name = attr.getAttribute("name", null);
    p.setAttribute("name", name);
    p.setAttribute("type", getAttributeType(attrControllerClass, attr, name));
    p.addSubElement(
      new XMLElement("<![CDATA[" + attr.getAttribute("value", null) + "]]>"));
    r.addSubElement(p);
  }

  /**
   * Returns the type of the given attribute.
   *
   * @param attrControllerClass the attribute controller class.
   * @param attr an 'attribute' element.
   * @param name the name of the <tt>attr</tt> attribute.
   * @return the type of the given attribute.
   * @throws ParserException if the type of the given element cannot be found.
   */

  private String getAttributeType (
    final Class attrControllerClass,
    final XMLElement attr,
    final String name) throws ParserException
  {
    try {
      Method getter = attrControllerClass.getMethod("get" + name, new Class[0]);
      Class typ = getter.getReturnType();
      String s;
      if (typ == Byte.TYPE) {
        s = "byte";
      } else if (typ == Character.TYPE) {
        s = "char";
      } else if (typ == Short.TYPE) {
        s = "short";
      } else if (typ == Integer.TYPE) {
        s = "int";
      } else if (typ == Long.TYPE) {
        s = "long";
      } else if (typ == Float.TYPE) {
        s = "float";
      } else if (typ == Double.TYPE) {
        s = "double";
      } else if (typ == Boolean.TYPE) {
        s = "boolean";
      } else if (typ == String.class) {
        s = "string";
      } else {
        throw new ParserException(attr, "Invalid attribute type: " + typ, null);
      }
      return s;
    } catch (NoSuchMethodException e) {
      throw new ParserException(attr, "No such attribute: " + name, null);
    }
  }

  // -------------------------------------------------------------------------
  // Utility class
  // -------------------------------------------------------------------------

  /**
   * Thrown when a Fractal template cannot be converted to a Kilim template with
   * the same inheritance relations.
   */

  static class ConversionException extends ParserException {

    /**
     * Constructs a new {@link ConversionException ConversionException} object.
     */

    public ConversionException () {
      super("", null);
    }
  }
}
