/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.factory.InstantiationException;

import org.objectweb.fractal.util.Fractal;

/**
 * A generic launcher to start a Fractal application from the name of a template
 * definition.
 */

public class Launcher {

  /**
   * The bootstrap parser component, manually created.
   */

  private static Parser PARSER;

  /**
   * Private constructor (uninstantiable class).
   */

  private Launcher () {
  }

  /**
   * Instantiates and starts the Fractal application described by the given
   * template name. Usage:
   * <p>
   * <tt>Launcher <i>template</i> [ <i>itf</i> ] </tt>
   * <p>
   * where <i>template</i> is the name of the template to be instantiated and
   * started, and <i>itf</i> is the name of its Runnable interface.
   * <p>
   * This launcher initializes Fractal, constructs a composite parser component,
   * uses it to parse the given template definition, instantiates the template
   * component that just been parsed, starts the resulting component if it has
   * a LifeCycleController interface, and starts a new thread to launch the
   * application if the component provides a "run" interface of type Runnable.
   *
   * @param args the command line arguments.
   * @throws Exception if a problem occurs.
   */

  public static void main (final String[] args) throws Exception {
    // do initialization in a separate method call in order to not keep
    // references on unused objects (so that they can be collected).
    Component comp = init(args);
    // starts the application, if it provides a "run" interface
    Runnable run = null;
    try {
      run = (Runnable)comp.getFcInterface(args.length > 1 ? args[1] : "run");
    } catch (NoSuchInterfaceException ignored) {
    } catch (ClassCastException ignored) {
    }
    if (run != null) {
      run.run();
    }
  }

  /**
   * Instantiates the Fractal application described by the given template name.
   *
   * @param args the name of template to be instantiated.
   * @return the instantiated component.
   * @throws Exception if the component cannot be instantiated.
   */

  private static Component init (final String[] args) throws Exception {
    if (args.length < 1) {
      System.out.println("Usage: Launcher <template> [ <itf> ]");
      System.out.print("where <template> is the name of the template to be ");
      System.out.print("instantiated and started,\nand <itf> is the name of ");
      System.out.println("its Runnable interface");
      System.exit(1);
    }

    // creates the template and instantiates it
    Parser parser = getBootstrapParser();
    Component comp = parser.loadTemplate(args[0], false);

    // starts the application if it has a lifecycle controller
    LifeCycleController lc = null;
    try {
      lc = Fractal.getLifeCycleController(comp);
    } catch (NoSuchInterfaceException ignored) {
    }
    if (lc != null) {
      lc.startFc();
    }
    return comp;
  }

  /**
   * Returns the bootstrap parser component. This component is created directly
   * with the Fractal API (or even without it), since there is no pre existing
   * parser component to create it from an XML template descriptor. Multiple
   * calls to this method return the same component.
   *
   * @return the bootstrap parser component.
   * @throws InstantiationException if the bootstrap parser cannot be created.
   * @throws NoSuchInterfaceException if the bootstrap type factory or generic
   *      factory interface cannot be found.
   */

  public static Parser getBootstrapParser ()
    throws InstantiationException, NoSuchInterfaceException
  {
    if (PARSER == null) {
      Component boot = Fractal.getBootstrapComponent();
      // manual assembly of the parser components
      BasicLoader bl = new BasicLoader();
      ResolvingLoader rl = new ResolvingLoader();
      rl.bindFc("delegate", bl);
      BasicParser bp = new BasicParser();
      bp.bindFc("loader", rl);
      bp.bindFc("type-factory", Fractal.getTypeFactory(boot));
      bp.bindFc("generic-factory", Fractal.getGenericFactory(boot));
      PARSER = bp;
    }
    return PARSER;
  }
}
