/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.type.ComponentType;

/**
 * A component interface to load template components.
 */

public interface Parser {

  /**
   * Loads the component type whose name is given. Depending on the
   * implementation, this method can return or not a different object when
   * it is called several times with the same argument.
   *
   * @param name a component type name.
   * @return the component type object whose name is given.
   * @throws ParserException if the component type definition cannot be loaded,
   *      if it is invalid, or if the corresponding component type object cannot
   *      be created.
   */

  ComponentType loadType (String name) throws ParserException;

  /**
   * Loads the template component whose name is given. Depending on the
   * implementation, this method can return or not a different component when
   * it is called several times with the same argument.
   *
   * @param name a template name.
   * @param template <tt>true</tt> to return a template component,
   *      <tt>false</tt> to return directly the instantiated component.
   * @return the Component interface of the loaded template component.
   * @throws ParserException if the template definition cannot be loaded, if it
   *      is invalid, or if the corresponding template component cannot be
   *      created.
   */

  Component loadTemplate (String name, boolean template) throws ParserException;

  /**
   * Loads the template component whose name is given. This method always
   * returns a new component instance.
   *
   * @param name a template name.
   * @param template <tt>true</tt> to return a template component,
   *      <tt>false</tt> to return directly the instantiated component.
   * @return the Component interface of the loaded template component.
   * @throws ParserException if the template definition cannot be loaded, if it
   *      is invalid, or if the corresponding template component cannot be
   *      created.
   */

  Component createTemplate (String name, boolean template) throws ParserException;

  /**
   * Registers an external component under the given name. External components
   * are used to create the template components corresponding to
   * &lt;external-template&gt; template descriptors.
   *
   * @param name the name of the external component.
   * @param component an already instantiated component.
   */

  void registerExternalComponent (String name, Component component);
}
