/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import java.io.PrintStream;
import java.io.PrintWriter;

/**
 * Thrown when loading, parsing, checking or constructing a type or a template
 * definition.
 */

public class ParserException extends Exception {

  /**
   * The name of the file where this exception occured. May be <tt>null</tt>.
   */

  private final String src;

  /**
   * The line number in <tt>src</tt> where this exception occured.
   */

  private final int line;

  /**
   * The exception wrapped in this exception. May be <tt>null</tt>.
   */

  private final Throwable exception;

  // -------------------------------------------------------------------------
  // Constructors
  // -------------------------------------------------------------------------

  /**
   * Constructs a new {@link ParserException ParserException}.
   *
   * @param src the file where this exception occured. May be <tt>null</tt>.
   * @param line the line number where this exception occured.
   * @param msg a detail error message.
   * @param exception a wrapped exception. May be <tt>null</tt>.
   */

  public ParserException (
    final String src,
    final int line,
    final String msg,
    final Throwable exception)
  {
    super(msg);
    this.src = src;
    this.line = line;
    this.exception = exception;
  }

  /**
   * Constructs a new {@link ParserException ParserException}.
   *
   * @param src the XML element where this exception occured.
   * @param msg a detail error message.
   * @param exception a wrapped exception. May be <tt>null</tt>.
   */

  public ParserException (
    final XMLElement src,
    final String msg,
    final Throwable exception)
  {
    super(msg);
    this.src = src.source;
    this.line = src.line;
    this.exception = exception;
  }

  /**
   * Constructs a new {@link ParserException ParserException}.
   *
   * @param src the XML element where this exception occured.
   * @param exception a wrapped parser exception.
   */

  public ParserException (
    final XMLElement src,
    final ParserException exception)
  {
    super(exception.getMessage());
    this.src = (exception.src != null ? exception.src : src.source);
    this.line = exception.line;
    this.exception = exception.exception;
  }

  /**
   * Constructs a new {@link ParserException ParserException}.
   *
   * @param msg a detail error message.
   * @param exception a wrapped exception. May be <tt>null</tt>.
   */

  public ParserException (final String msg, final Throwable exception) {
    super(msg);
    this.src = null;
    this.line = 0;
    this.exception = exception;
  }

  // -------------------------------------------------------------------------
  // Accessors
  // -------------------------------------------------------------------------

  /**
   * Returns the name of the file where this exception occured.
   *
   * @return the name of the file where this exception occured. May be
   *      <tt>null</tt>.
   */

  public String getSource () {
    return src;
  }

  /**
   * Returns the line number in {@link #getSource getSource} where this
   * exception occured.
   *
   * @return the line number in {@link #getSource getSource} where this
   *      exception occured.
   */

  public int getLine () {
    return line;
  }

  /**
   * Returns the exception that is wrapped in this exception.
   *
   * @return the exception wrapped in this exception. May be <tt>null</tt>.
   */

  public Throwable getException () {
    return exception;
  }

  // -------------------------------------------------------------------------
  // Overriden methods
  // -------------------------------------------------------------------------

  /**
   * Returns a String representation of this exception.
   *
   * @return a String representation of this exception.
   */

  public String toString () {
    StringBuffer buf = new StringBuffer();
    buf.append("ERROR");
    if (src != null) {
      buf.append(" at ");
      buf.append(src);
      buf.append(", line ");
      buf.append(line);
    }
    buf.append(":\n  ");
    buf.append(getMessage());
    return buf.toString();
  }

  /**
   * Prints the stack backtrace.
   */

  public void printStackTrace () {
    if (exception != null) {
      exception.printStackTrace();
    }
    super.printStackTrace();
  }
  
  /**
   * Prints this exception and its backtrace to the specified print stream.
   *
   * @param s <tt>PrintStream</tt> to use for output.
   */

  public void printStackTrace (final PrintStream s) {
    if (exception != null) {
      exception.printStackTrace(s);
    }
    super.printStackTrace(s);
  }

  /**
   * Prints this exception and its backtrace to the specified print writer.
   *
   * @param s <tt>PrintWriter</tt> to use for output.
   */

  public void printStackTrace (final PrintWriter s) {
    if (exception != null) {
      exception.printStackTrace(s);
    }
    super.printStackTrace(s);
  }
}
