/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.objectweb.util.monolog.api.BasicLevel;

import java.util.HashMap;

/**
 * A {@link CachingLoader CachingLoader} that resolves the inheritance
 * relations between definitions. This loader transforms definitions that extend
 * other definitions into equivalent definitions that do not use inheritance.
 * This loader also checks typing constraints on the definitions it loads.
 */

public class ResolvingLoader extends CachingLoader {

  /**
   * Gives the name of the attribute to be used to identify sub elements of a
   * given element, for the XML merging algorithm. If this name is <tt>null</tt>
   * for an element 'foo', then two 'foo' elements are merged by recursively
   * merging their sub elements that have the same tag. If this name is 'attr',
   * then two elements are merged by merging their set of sub elements, by using
   * the value of the 'attr' attribute of these sub elements to distinguish
   * them.
   */

  private final static HashMap IDENTIFIER_ATTR;

  /**
   * Gives the order of the sub elements of a given element, as an array of
   * String.
   */

  private final static HashMap SUBELEMENTS_ORDER;

  /**
   * Initialize the static fields of this class.
   */

  static {
    IDENTIFIER_ATTR = new HashMap();
    IDENTIFIER_ATTR.put("provides", "name");
    IDENTIFIER_ATTR.put("requires", "name");
    IDENTIFIER_ATTR.put("attributes", "name");
    IDENTIFIER_ATTR.put("components", "name");
    IDENTIFIER_ATTR.put("bindings", "client");
    IDENTIFIER_ATTR.put("sharing", "path");

    SUBELEMENTS_ORDER = new HashMap();
    SUBELEMENTS_ORDER.put(
      "component-type", new String[] {
        "provides", "requires" });
    SUBELEMENTS_ORDER.put(
      "primitive-template", new String[] {
        "primitive-content", "controller" });
    SUBELEMENTS_ORDER.put(
      "composite-template", new String[] {
        "composite-content", "controller" });
    SUBELEMENTS_ORDER.put(
      "controller", new String[] {
        "attributes", "template-controller", "component-controller" });
    SUBELEMENTS_ORDER.put(
      "composite-content", new String[] {
        "components", "bindings", "sharing" });
  }

  /**
   * Indicates if generated attributes must be added to the resolved elements.
   * See {@link #ResolvingLoader(boolean) ResolvingLoader}.
   */

  private boolean comment;

  // -------------------------------------------------------------------------
  // Constructors
  // -------------------------------------------------------------------------

  /**
   * Constructs a new {@link ResolvingLoader ResolvingLoader} object.
   */

  public ResolvingLoader () {
  }

  /**
   * Constructs a new {@link ResolvingLoader ResolvingLoader} object.
   *
   * @param comment <tt>true</tt> to add generated attributes to the resolved
   *      elements to indicate the source of each attribute or element (this
   *      source can be the base element, the super element, or a merged of
   *      these two elements. The generated attribute is empty in the first
   *      case, <tt>[src]="super"</tt> in the second case, and
   *      <tt>[src]="merge"</tt> in the last case - the attribute name is
   *      voluntarily not a valid identifier to show that this attribute has
   *      been generated, and to avoid conflicts with normal attributes).
   */

  public ResolvingLoader (final boolean comment) {
    this.comment = comment;
  }

  // -------------------------------------------------------------------------
  // Implementation of the abstract inherited methods
  // -------------------------------------------------------------------------

  protected XMLElement handleTypeDescriptor (
    final String name,
    final XMLElement desc) throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "resolving type descriptor: " + name);
    }
    return resolve(desc, false);
  }

  protected XMLElement handleTemplateDescriptor (
    final String name,
    final XMLElement desc) throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "resolving template descriptor: " + name);
    }
    return resolve(desc, true);
  }

  // -------------------------------------------------------------------------
  // Utility methods
  // -------------------------------------------------------------------------

  /**
   * Resolves the given type or template definition.
   *
   * @param desc a type or template definition.
   * @param template <tt>true</tt> if the given definition is a template
   *      definition, or <tt>false</tt> if it is a type definition.
   * @return a type or template definition equivalent to the given one, but that
   *      does not use inheritance.
   * @throws ParserException if a super definition cannot be loaded, or if the
   *      resolution fails.
   */

  private XMLElement resolve (XMLElement desc, final boolean template)
    throws ParserException
  {
    String extend = desc.getAttribute("extends", "");
    if (extend.length() > 0) {
      // makes a deep copy of desc, in order not to modify the original element
      desc = new XMLElement(desc);
      // removes the extend attribute
      if (!comment) {
        desc.setAttribute("extends", null);
      }
      // loads the inherited definition
      XMLElement superDesc;
      if (template) {
        superDesc = loadTemplateDescriptor(extend);
      } else {
        superDesc = loadTypeDescriptor(extend);
      }
      // checks the definition and its inherited definitions have the same tag
      if (!superDesc.tag.equals(desc.tag)) {
        throw new ParserException(
          "A " + desc.tag + " cannot extend a " + superDesc.tag, null);
      }
      // merges the definition and its inherited definition
      merge(desc, superDesc, superDesc.source, superDesc.line);
    }
    return desc;
  }

  /**
   * Merges the given definition and the given inherited definition. This method
   * merges the attributes of the two definitions, and recursively calls itself
   * to merges the sub elements of the two elements.
   *
   * @param elem a definition.
   * @param superElem the definition extended by <tt>elem</tt>.
   * @param source the file from which the root element to which superElem
   *      belongs was parsed.
   * @param line the line from which the root element to which superElem
   *      belongs was parsed.
   * @throws ParserException if a problem occurs.
   */

  private void merge (
    final XMLElement elem,
    final XMLElement superElem,
    final String source,
    final int line) throws ParserException
  {
    if (comment) {
      elem.setAttribute("[src]", "merge");
    }
    // merge attributes
    String[] attrs = superElem.getAttributes();
    for (int i = 0; i < attrs.length; ++i) {
      String value = elem.getAttribute(attrs[i], "");
      String superValue = superElem.getAttribute(attrs[i], null);
      if (value.length() == 0) {
        elem.setAttribute(attrs[i], superValue);
      } else if (comment && !value.equals(superValue)) {
        elem.setAttribute("[" + attrs[i] + "]", superValue);
      }
    }
    // merge sub elements
    String nameAttr = (String)IDENTIFIER_ATTR.get(elem.tag);

    // if nameAttr is null the sub elements are merged in the following way:
    //   <elem>              <superElem>
    //     <b> ... </b>        <a> ... </a>
    //     <c> ... </c>        <b> ... </b>
    //   </elem>             </superElem>
    // gives:
    //   <elem>
    //     <a> ... </a> <!-- inherited from superElem as is -->
    //     <b> ... </b> <!-- recursive merging of the b elements -->
    //     <c> ... </c> <!-- copied from elem as is -->
    //   </elem>

    // if nameAttr is 'attr' the sub elements are merged in the following way:
    //   <elem>                         <superElem>
    //     <a attr="2"> ... </a>          <a attr="1"> ... </a>
    //     <a attr="3"> ... </a>          <a attr="2"> ... </a>
    //   </elem>                        </superElem>
    // gives:
    //   <elem>
    //     <a attr="1"> ... </a> <!-- inherited from superElem as is -->
    //     <a attr="2"> ... </a> <!-- recursive merging of the "2" elements -->
    //     <a attr="3"> ... </a> <!-- copied from elem as is -->
    //   </elem>

    for (int i = 0; i < superElem.getSubElementCount(); ++i) {
      XMLElement superNode = superElem.getSubElement(i);
      String superName;
      if (nameAttr == null) {
        superName = superNode.tag;
      } else {
        superName = superNode.getAttribute(nameAttr, null);
      }
      XMLElement node = null;
      for (int j = 0; j < elem.getSubElementCount(); ++j) {
        XMLElement n = elem.getSubElement(j);
        String name;
        if (nameAttr == null) {
          name = n.tag;
        } else {
          name = n.getAttribute(nameAttr, null);
        }
        if (name.equals(superName)) {
          node = n;
          break;
        }
      }
      if (node == null) {
        // makes a deep copy of superNode, in order not to modify it directly
        superNode = new XMLElement(superNode);
        elem.addSubElement(superNode);
      } else {
        merge(node, superNode, source, line);
      }
    }
    if (nameAttr == null) {
      int n = elem.getSubElementCount();
      if (n > 0) {
        // sorts the sub elements of 'elem'
        String[] order = (String[])SUBELEMENTS_ORDER.get(elem.tag);
        for (int i = 0; i < order.length; ++i) {
          XMLElement e = elem.getSubElement(order[i]);
          if (e != null) {
            elem.addSubElement(e);
          }
        }
        for (int i = 0; i < n; ++i) {
          elem.removeSubElement(0);
        }
      }
    }
  }
}
