/*
 * Simple Fractal ADL Parser
 * Copyright (C) 2003 Ecole des Mines de Nantes / INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Pierre-Charles.David@emn.fr
 *
 * Author: Eric Bruneton, Pierre-Charles David
 */

header { package org.objectweb.fractal.adl; }
{
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.objectweb.fractal.adl.XMLElement;
}

// ---------------------------------------------------------------------------
// PARSER
// ---------------------------------------------------------------------------

class SimpleADLParser extends Parser ;
options {
    k = 1;
}

{
    /**
     * Creates a <code>&lt;component-type&gt;</code> element.
     *
     * @param name the name to give to the type (<tt>name</tt> attribute)
     * @param extend the name of the supertype (<tt>extends</tt> attribute)
     * @param provides a list of <code>&lt;interface-type&gt;</code>
     * <code>XMLElement</code>s corresponding to the server interfaces provided
     * by the component type to create
     * @param provides a list of <code>&lt;interface-type&gt;</code>
     * <code>XMLElement</code>s corresponding to the client interfaces required
     * by the component type to create
     */
    private XMLElement createComponentType(String name, String extend, List provides, List requires) {
        XMLElement elem = new XMLElement("component-type");
        elem.setAttribute("name", name);
        if (extend != null) {
            elem.setAttribute("extends", extend);
        }
        if (provides != null && provides.size() > 0) {
            XMLElement provideElem = new XMLElement("provides");
            for (int i = 0; i < provides.size(); ++i) {
                provideElem.addSubElement((XMLElement)provides.get(i));
            }
            elem.addSubElement(provideElem);
        }
        if (requires != null && requires.size() > 0) {
            XMLElement requireElem = new XMLElement("requires");
            for (int i = 0; i < requires.size(); ++i) {
                requireElem.addSubElement((XMLElement)requires.get(i));
            }
            elem.addSubElement(requireElem);
        }
        return elem;
    }

    /**
     * Creates a <code>&lt;controller&gt;</code> element.
     *
     * @param tmplDesc a <code>&lt;template-controller&gt;</code>
     * @param compDesc a <code>&lt;component-controller&gt;</code>
     */
    private XMLElement createControllerDescriptor(XMLElement tmplDesc, XMLElement compDesc) {
        if (tmplDesc != null || compDesc != null) {
            XMLElement controller = new XMLElement("controller");
            if (tmplDesc != null) {
                controller.addSubElement(tmplDesc);
            }
            if (compDesc != null) {
                controller.addSubElement(compDesc);
            }
            return controller;
        } else {
            return null;
        }
    }

    /**
     * Finds the name of the declared type of a component declaration.
     *
     * @param decl the list of existing components and type declarations (as
     * <code>XMLElement</code>s)
     * @param componentName the name of the component which type is looked for
     */
    private String getComponentTypeName(List decl, String componentName) {
        for (Iterator iter = decl.iterator(); iter.hasNext(); ) {
            XMLElement elt = (XMLElement) iter.next();
            try {
                if (componentName.equals(elt.getAttribute("name", ""))) {
                    return elt.getAttribute("implements", "");
                }
            } catch (ParserException pe) {
                pe.printStackTrace();
            }
        }
        return null;
    }

}

/**
 * Parses an <code>architecture.adl</code> file and returns a list of equivalent
 * <code>XMLElement</code> declarations (conforming to Fractal ADL's DTD).
 */
architecture returns [List declarations  = new ArrayList()]
    :   ( typeDeclaration[declarations]
        | primitiveDeclaration[declarations]
        | compositeDeclaration[declarations]
        )*
    ;

// TYPE DECLARATION ----------------------------------------------------------

typeDeclaration [List decl]
{
    String typeName = null;
    String superType = null;
    List provided = new ArrayList();
    List required = new ArrayList();
}
    :
        "type" typeName = qualifiedName ( "extends" superType = qualifiedName )?
        LCURLY
        ( "provides" interfaceType[provided]
        | "requires" interfaceType[required] )*
        RCURLY
        {
            decl.add(createComponentType(typeName, superType, provided, required));
        }
    ;

interfaceType [List itfs]
{
    String name = null;
    String signature = null;
    boolean mandatory = true;
    boolean single = true;
}
:
   mandatory = contingency single = cardinality signature = qualifiedName name = qualifiedName SEMI
        {
            XMLElement elem = new XMLElement("interface-type");
            elem.setAttribute("name", name);
            elem.setAttribute("signature", signature);
            if (!mandatory) {
                elem.setAttribute("contingency", "optional");
            }
            if (!single) {
                elem.setAttribute("cardinality", "collection");
            }
            itfs.add(elem);
        }
;

contingency returns [boolean mandatory = true]
    :
        ( "optional"  { mandatory = false; }
        | "mandatory" { mandatory = true; }
        )?
    ;

cardinality returns [boolean single = true]
    :
        ( "collection" { single = false; }
        | "single"     { single = true; }
        )?
    ;

qualifiedName returns [String value = null]
{
  StringBuffer buf = new StringBuffer();
}
:
  i:IDENT
  { buf.append(i.getText()); }
  (
    DOT j:IDENT
    { buf.append(".").append(j.getText()); }
  )*
  {
    value = buf.toString();
  }
;

// PRIMITIVE COMPONENT -------------------------------------------------------

primitiveDeclaration [List decl]
{
    String name = null;
    String className = null;
    String typeName = null;
    XMLElement primitive = new XMLElement("primitive-template");
    XMLElement tmplDesc = null;
    XMLElement compDesc = null;
    List provided = new ArrayList();
    List required = new ArrayList();
}
    : "primitive" name = qualifiedName "is" className = qualifiedName
        ( "implements" typeName = qualifiedName )?
        ( ( LCURLY
            ( "provides" interfaceType[provided]
            | "requires" interfaceType[required]
            | tmplDesc = templateController
            | compDesc = componentController
            )*
            RCURLY
          )
        | SEMI
        )

        {
            primitive.setAttribute("name", name);
            if (provided.size() > 0 || required.size() > 0) {
                decl.add(createComponentType(name + "Type", typeName, provided, required));
                typeName = name + "Type";
            }
            primitive.setAttribute("implements", typeName);
            XMLElement controller = createControllerDescriptor(tmplDesc, compDesc);
            if (controller != null) {
                primitive.addSubElement(controller);
            }
            XMLElement content = new XMLElement("primitive-content");
            content.setAttribute("class", className);
            primitive.addSubElement(content);
            decl.add(primitive);
        }
;

// COMPOSITE COMPONENT -------------------------------------------------------

compositeDeclaration [List decl]
{
    String name = null;
    String typeName = null;
    XMLElement composite = new XMLElement("composite-template");
    XMLElement tmplDesc = null;
    XMLElement compDesc = null;
    List provided = new ArrayList();
    List required = new ArrayList();
    XMLElement bindings = new XMLElement("bindings");
    XMLElement subcomponents = new XMLElement("components");
}
    : "composite" name = qualifiedName ("implements" typeName = qualifiedName )?
        LCURLY
        ( "provides" interfaceType[provided]
        | "requires" interfaceType[required]
        | subComponent[decl, subcomponents]
        | binding[bindings]
        | tmplDesc = templateController
        | compDesc = componentController
        )*
        RCURLY
        {
            composite.setAttribute("name", name);
            if (provided.size() > 0 || required.size() > 0) {
                decl.add(createComponentType(name + "Type", typeName, provided, required));
                typeName = name + "Type";
            }
            composite.setAttribute("implements", typeName);
            XMLElement content = new XMLElement("composite-content");
            content.addSubElement(subcomponents);
            content.addSubElement(bindings);
            composite.addSubElement(content);

            XMLElement controller = createControllerDescriptor(tmplDesc, compDesc);
            if (controller != null) {
                composite.addSubElement(controller);
            }
            decl.add(composite);
        }
    ;

subComponent[List decl, XMLElement components]
{
    String name = null;
}
    : "contains" name = qualifiedName "as" localName:IDENT SEMI
        {
            XMLElement self = new XMLElement("component");
            self.setAttribute("name", localName.getText());
            self.setAttribute("implementation", name);
            String typeName = getComponentTypeName(decl, name);
            if (typeName != null) {
                self.setAttribute("type", typeName);
            }
            components.addSubElement(self);
        }
    ;

binding [XMLElement bindings]
    : "binds" cltComp:IDENT DOT cltItf:IDENT "to" srvComp:IDENT DOT srvItf:IDENT SEMI
        {
            XMLElement self = new XMLElement("binding");
            self.setAttribute("client", cltComp.getText() + "." + cltItf.getText());
            self.setAttribute("server", srvComp.getText() + "." + srvItf.getText());
            bindings.addSubElement(self);
        }
    ;

templateController returns [XMLElement self = null]
    : "template-controller" "is" id:IDENT SEMI
        {
            self = new XMLElement("template-controller");
            self.setAttribute("desc", id.getText());
        }
    ;

componentController returns [XMLElement self = null]
    : "component-controller" "is" id:IDENT SEMI
        {
            self = new XMLElement("component-controller");
            self.setAttribute("desc", id.getText());
        }
    ;

// ---------------------------------------------------------------------------
// LEXER
// ---------------------------------------------------------------------------

class SimpleADLLexer extends Lexer;

options {
    k=4;
    charVocabulary='\u0003'..'\uFFFF';
}

// OPERATORS
LCURLY    :  '{' ;
RCURLY    :  '}' ;
DOT       :  '.' ;
SEMI      :  ';' ;

// Whitespace -- ignored
WS  :  (  ' '
        |  '\t'
        |  '\f'
            // handle newlines
        |  (  options {generateAmbigWarnings=false;}
            :  "\r\n"  // Evil DOS
            |  '\r'    // Macintosh
            |  '\n'    // Unix (the right way)
            )
            { newline(); }
        )+
        { _ttype = Token.SKIP; }
    ;

// Single-line comments
SL_COMMENT
    :  "//"
        (~('\n'|'\r'))* ('\n'|'\r'('\n')?)
        {$setType(Token.SKIP); newline();}
    ;

// Identifiers
IDENT
    :  ('a'..'z'|'A'..'Z'|'_'|'$') ('a'..'z'|'A'..'Z'|'_'|'$'|'0'..'9'|'-')*
    ;
