/***
 * Simple Fractal ADL Loader
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.objectweb.fractal.api.control.BindingController;

import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.io.InputStream;
import java.io.InputStreamReader;

import java.util.List;
import java.util.Map;
import java.util.HashMap;

/**
 * A loader for an alternative parser which understands a simplified,
 * more user-friendly ADL syntax.
 */
public class SimpleADLLoader implements Loader, BindingController {

  /**
   * The optional logger factory used to get a logger for this component.
   */

  protected LoggerFactory loggerFactory;

  /**
   * The logger used to log messages. May be <tt>null</tt>.
   */

  private Logger logger;

  /**
   * The type and template definitions converted from the adl file.
   */
  
  private Map definitions;

  // -------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // -------------------------------------------------------------------------

  public String[] listFc () {
    return new String[] { "logger-factory" };
  }

  public Object lookupFc (final String clientItfName) {
    if (clientItfName.equals("logger-factory")) {
      return loggerFactory;
    }
    return null;
  }

  public void bindFc (final String clientItfName, final Object serverItf) {
    if (clientItfName.equals("logger-factory")) {
      loggerFactory = (LoggerFactory)serverItf;
      logger = loggerFactory.getLogger(getClass().getName());
    }
  }

  public void unbindFc (final String clientItfName) {
    if (clientItfName.equals("logger-factory")) {
      loggerFactory = null;
      logger = null;
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the Loader interface
  // -------------------------------------------------------------------------

  public XMLElement loadTypeDescriptor (final String name)
    throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "loading type descriptor: " + name);
    }
    return load(name);
  }

  public XMLElement loadTemplateDescriptor (final String name)
    throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "loading template descriptor: " + name);
    }
    return load(name);
  }

  // -------------------------------------------------------------------------
  // Utility methods
  // -------------------------------------------------------------------------

  /**
   * Loads the definition whose name is given.
   *
   * @param name the name of type or template definition.
   * @return the XML element corresponding to the specified definition.
   * @throws ParserException if the definition cannot be found.
   */
  private XMLElement load (final String name) throws ParserException {
    if (definitions == null) {
      definitions = new HashMap();
      InputStream is = ClassLoader.getSystemClassLoader().getResourceAsStream("architecture.adl");
      if (is == null) {
        throw new ParserException("Cannot find the 'architecture.adl' file", null);
      }
      SimpleADLLexer lexer = new SimpleADLLexer(new InputStreamReader(is));
      SimpleADLParser parser = new SimpleADLParser(lexer);
      List list;
      try {
        list = parser.architecture();
      } catch (Throwable e) {
        throw new ParserException("Error in 'architecture.adl'", e);
      }
      for (int i = 0; i < list.size(); ++i) {
        XMLElement elem = (XMLElement)list.get(i);
// DEBUG
//        try {
//          elem.write(new java.io.PrintWriter(System.err, true));
//        } catch (IOException e) { }
// END DEBUG
        definitions.put(elem.getAttribute("name", null), elem);
      }
    }
    XMLElement result = (XMLElement)definitions.get(name);
    if (result == null) {
      throw new ParserException(name + " not found in 'architecture.adl'", null);
    }
    return result;
  }
}
