/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * A {@link DelegatingLoader DelegatingLoader} that verifies that the loaded
 * elements conform to the Fractal ADL DTD.
 */

public class VerifyingLoader extends DelegatingLoader {

  // -------------------------------------------------------------------------
  // Implementation of the abstract inherited methods
  // -------------------------------------------------------------------------

  protected XMLElement handleTypeDescriptor (
    final String name,
    final XMLElement desc) throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "verifying type descriptor: " + name);
    }
    checkComponentType(desc);
    return desc;
  }

  protected XMLElement handleTemplateDescriptor (
    final String name,
    final XMLElement desc) throws ParserException
  {
    if (logger != null && logger.isLoggable(BasicLevel.INFO)) {
      logger.log(BasicLevel.INFO, "verifying template descriptor: " + name);
    }
    if (desc.tag.equals("primitive-template")) {
      checkPrimitiveTemplate(desc);
    } else if (desc.tag.equals("composite-template")) {
      checkCompositeTemplate(desc);
    } else if (desc.tag.equals("external-template")) {
      checkExternalTemplate(desc);
    } else {
      throw new ParserException(
        null,
        desc.line,
        "'primitive-', 'composite-' or 'external-template' tag expected",
        null);
    }
    return desc;
  }

  // -------------------------------------------------------------------------
  // Utility methods
  // -------------------------------------------------------------------------

  // component types ---------------------------------------------------------

  /**
   * Checks a 'component-type' element.
   *
   * @param type the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkComponentType (final XMLElement type)
    throws ParserException
  {
    if (!type.tag.equals("component-type")) {
      throw new ParserException(
        null, type.line, "'component-type' tag expected", null);
    }
    type.checkAttributes("name,extends");
    type.getAttribute("name", null);
    type.checkSubElements("provides?,requires?");
    for (int i = 0; i < type.getSubElementCount(); ++i) {
      XMLElement elem = type.getSubElement(i);
      checkInterfaceTypes(elem);
    }
  }

  /**
   * Checks a 'provides' or 'requires' element.
   *
   * @param types the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkInterfaceTypes (final XMLElement types)
    throws ParserException
  {
    types.checkAttributes("");
    types.checkSubElements("interface-type*");
    for (int i = 0; i < types.getSubElementCount(); ++i) {
      XMLElement elem = types.getSubElement(i);
      checkInterfaceType(elem);
    }
  }

  /**
   * Checks an 'interface-type' element.
   *
   * @param type the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkInterfaceType (final XMLElement type)
    throws ParserException
  {
    type.checkAttributes("name,signature,contingency,cardinality");
    type.getAttribute("name", null);
    String contingency = type.getAttribute("contingency", "mandatory");
    if (!contingency.equals("mandatory") && !contingency.equals("optional")) {
      throw new ParserException(
        null, type.line, "Invalid contingency value: " + contingency, null);
    }
    String cardinality = type.getAttribute("cardinality", "single");
    if (!cardinality.equals("single") && !cardinality.equals("collection")) {
      throw new ParserException(
        null, type.line, "Invalid cardinality value: " + cardinality, null);
    }
    type.checkSubElements("");
  }

  // templates ---------------------------------------------------------------

  /**
   * Checks a 'composite-template' or 'primitive-template' element.
   *
   * @param tmpl the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkTemplate (final XMLElement tmpl)
    throws ParserException
  {
    tmpl.checkAttributes("name,implements,extends");
    tmpl.getAttribute("name", null);
    if (tmpl.getAttribute("extends", "").equals("")) {
      tmpl.getAttribute("implements", null);
    }
  }

  // primitive templates -----------------------------------------------------

  /**
   * Checks a 'primitive-template' element.
   *
   * @param tmpl the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkPrimitiveTemplate (final XMLElement tmpl)
    throws ParserException
  {
    checkTemplate(tmpl);
    tmpl.checkSubElements("primitive-content?,controller?");
    for (int i = 0; i < tmpl.getSubElementCount(); ++i) {
      XMLElement elem = tmpl.getSubElement(i);
      if (elem.tag.equals("primitive-content")) {
        checkPrimitiveContent(elem);
      } else {
        checkController(elem);
      }
    }
  }

  /**
   * Checks a 'primitive-content' element.
   *
   * @param content the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkPrimitiveContent (final XMLElement content)
    throws ParserException
  {
    content.checkAttributes("class");
    content.getAttribute("class", null);
    content.checkSubElements("");
  }

  // composite templates -----------------------------------------------------

  /**
   * Checks a 'composite-template' element.
   *
   * @param tmpl the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkCompositeTemplate (final XMLElement tmpl)
    throws ParserException
  {
    checkTemplate(tmpl);
    tmpl.checkSubElements("composite-content?,controller?");
    for (int i = 0; i < tmpl.getSubElementCount(); ++i) {
      XMLElement elem = tmpl.getSubElement(i);
      if (elem.tag.equals("composite-content")) {
        checkCompositeContent(elem);
      } else {
        checkController(elem);
      }
    }
  }

  /**
   * Checks a 'composite-content' element.
   *
   * @param content the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkCompositeContent (final XMLElement content)
    throws ParserException
  {
    content.checkAttributes("");
    content.checkSubElements("components?,bindings?,sharing?");
    for (int i = 0; i < content.getSubElementCount(); ++i) {
      XMLElement elem = content.getSubElement(i);
      if (elem.tag.equals("components")) {
        checkComponents(elem);
      } else if (elem.tag.equals("bindings")) {
        checkBindings(elem);
      } else if (elem.tag.equals("sharing")) {
        checkSharing(elem);
      }
    }
  }

  /**
   * Checks a 'components' element.
   *
   * @param comps the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkComponents (final XMLElement comps)
    throws ParserException
  {
    comps.checkAttributes("");
    comps.checkSubElements("component*");
    for (int i = 0; i < comps.getSubElementCount(); ++i) {
      XMLElement elem = comps.getSubElement(i);
      checkComponent(elem);
    }
  }

  /**
   * Checks a 'component' element.
   *
   * @param comp the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkComponent (final XMLElement comp)
    throws ParserException
  {
    comp.checkAttributes("name,type,implementation");
    comp.getAttribute("name", null);
    comp.checkSubElements("");
  }

  /**
   * Checks a 'bindings' element.
   *
   * @param bindings the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkBindings (final XMLElement bindings)
    throws ParserException
  {
    bindings.checkAttributes("");
    bindings.checkSubElements("binding*");
    for (int i = 0; i < bindings.getSubElementCount(); ++i) {
      XMLElement elem = bindings.getSubElement(i);
      checkBinding(elem);
    }
  }

  /**
   * Checks a 'binding' element.
   *
   * @param binding the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkBinding (final XMLElement binding)
    throws ParserException
  {
    binding.checkAttributes("client,server");
    binding.getAttribute("client", null);
    binding.getAttribute("server", null);
    binding.checkSubElements("");
  }

  /**
   * Checks a 'sharing' element.
   *
   * @param sharing the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkSharing (final XMLElement sharing)
    throws ParserException
  {
    sharing.checkAttributes("");
    sharing.checkSubElements("shared-component*");
    for (int i = 0; i < sharing.getSubElementCount(); ++i) {
      XMLElement elem = sharing.getSubElement(i);
      checkSharedComponent(elem);
    }
  }

  /**
   * Checks a 'shared-component' element.
   *
   * @param shared the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkSharedComponent (final XMLElement shared)
    throws ParserException
  {
    shared.checkAttributes("path,ref");
    shared.getAttribute("path", null);
    shared.getAttribute("ref", null);
    shared.checkSubElements("");
  }

  // external templates ------------------------------------------------------

  /**
   * Checks an 'external-template' element.
   *
   * @param tmpl the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkExternalTemplate (final XMLElement tmpl)
    throws ParserException
  {
    tmpl.checkAttributes("name,implements");
    tmpl.getAttribute("name", null);
    tmpl.getAttribute("implements", null);
    tmpl.checkSubElements("reference");
    XMLElement ref = tmpl.getSubElement(0);
    ref.checkAttributes("name");
    ref.getAttribute("name", null);
  }

  // controller definition ---------------------------------------------------

  /**
   * Checks a 'controller' element.
   *
   * @param cntrl the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkController (final XMLElement cntrl)
    throws ParserException
  {
    cntrl.checkAttributes("");
    cntrl.checkSubElements("attributes?,template-controller?,component-controller?");
    for (int i = 0; i < cntrl.getSubElementCount(); ++i) {
      XMLElement elem = cntrl.getSubElement(i);
      if (elem.tag.equals("attributes")) {
        checkAttributes(elem);
      } else {
        elem.checkAttributes("desc");
        elem.getAttribute("desc", null);
        elem.checkSubElements("");
      }
    }
  }

  // attributes --------------------------------------------------------------

  /**
   * Checks an 'attributes' element.
   *
   * @param attrs the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkAttributes (final XMLElement attrs)
    throws ParserException
  {
    attrs.checkAttributes("signature");
    attrs.getAttribute("signature", null);
    attrs.checkSubElements("attribute*");
    for (int i = 0; i < attrs.getSubElementCount(); ++i) {
      XMLElement elem = attrs.getSubElement(i);
      checkAttribute(elem);
    }
  }

  /**
   * Checks an 'attribute' element.
   *
   * @param attr the element to be checked.
   * @throws ParserException if the given element is invalid.
   */

  private void checkAttribute (final XMLElement attr)
    throws ParserException
  {
    attr.checkAttributes("name,value");
    attr.getAttribute("name", null);
    attr.getAttribute("value", null);
    attr.checkSubElements("");
  }
}
