/***
 * Fractal ADL Parser
 * Copyright (C) 2002 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl;

import java.io.EOFException;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

/**
 * An XML element. This class can also parse XML files, and can pretty print
 * XML elements. It also offer some methods to help manually validate XML
 * elements. But it does not handle XML entities, PCDATA and CDATA
 * elements, XML DTDs, namespaces...
 */

public class XMLElement {

  /**
   * The tag of this XML element.
   */

  public String tag;

  /**
   * The attributes of this XML element.
   */

  private HashMap attributes;

  /**
   * The sub elements of this XML element.
   */

  private ArrayList subElements;

  /**
   * The file from which this XML element was parsed. May be <tt>null</tt>.
   */

  public String source;

  /**
   * The line number where this XML element starts.
   */

  public int line;

  // --------------------------------------------------------------------------

  /**
   * The reader from which an XML element must be parsed.
   */

  private static LineNumberReader r;

  /**
   * The current character.
   */

  private static char c;

  /**
   * The current line.
   */

  private static int l;

  // --------------------------------------------------------------------------
  // Constructors
  // --------------------------------------------------------------------------

  /**
   * Creates a new empty XML element.
   *
   * @param tag the tag of the element.
   */

  public XMLElement (final String tag) {
    this.tag = tag;
  }

  /**
   * Creates a deep copy of the given XML element.
   *
   * @param elem the element to be cloned.
   */

  public XMLElement (final XMLElement elem) {
    tag = elem.tag;
    if (elem.attributes != null) {
      attributes = new HashMap(elem.attributes);
    }
    if (elem.subElements != null) {
      subElements = new ArrayList(elem.subElements);
      for (int i = 0; i < subElements.size(); ++i) {
        subElements.set(i, new XMLElement((XMLElement)subElements.get(i)));
      }
    }
    source = elem.source;
    line = elem.line;
  }

  // --------------------------------------------------------------------------
  // Accessor methods
  // --------------------------------------------------------------------------

  /**
   * Returns the names of the attributes of this XMLElement.
   *
   * @return the names of the attributes of this XMLElement.
   */

  public final String[] getAttributes () {
    if (attributes == null) {
      return new String[0];
    } else {
      String[] keys = new String[attributes.size()];
      Iterator iter = attributes.keySet().iterator();
      int i = 0;
      while (iter.hasNext()) {
        keys[i++] = (String)iter.next();
      }
      return keys;
    }
  }

  /**
   * Returns an attribute of this XML element.
   *
   * @param key the name of an attribute of this element.
   * @param defaultValue the default value of this attribute, or <tt>null</tt>
   *      if there is no default value.
   * @return the value of the given attribute of this element.
   * @throws ParserException if this element does not have such an attribute,
   *      and the default value is <tt>null</tt>.
   * @see #setAttribute(String, String)
   */

  public final String getAttribute (
    final String key,
    final String defaultValue) throws ParserException
  {
    String value = (attributes == null ?  null : (String)attributes.get(key));
    if (value == null) {
      if (defaultValue == null) {
        throw checkErr("\'" + key + "' attribute missing");
      }
      return defaultValue;
    }
    return value;
  }

  /**
   * Sets an attribute of this XML element.
   *
   * @param key the name of an attribute of this element.
   * @param value the value of this attribute. May be <tt>null</tt> to remove
   *      the attribute.
   * @see #getAttribute(String, String)
   */

  public final void setAttribute (final String key, final String value) {
    if (attributes == null) {
      attributes = new HashMap();
    }
    if (value == null) {
      attributes.remove(key);
    } else {
      attributes.put(key, value);
    }
  }

  /**
   * Returns the number of sub elements of this XML element.
   *
   * @return the number of sub elements of this XML element.
   */

  public final int getSubElementCount () {
    return (subElements == null ? 0 : subElements.size());
  }

  /**
   * Returns the sub element of this XML element whose index is given.
   *
   * @param index the index of a sub element of this XML element.
   * @return the sub element of this XML element whose index is given.
   */

  public final XMLElement getSubElement (final int index) {
    if (subElements == null) {
      throw new ArrayIndexOutOfBoundsException(index);
    }
    return (XMLElement)subElements.get(index);
  }

  /**
   * Returns the first sub element of this XML element whose tag is equal to the
   * given tag.
   *
   * @param tag a tag name.
   * @return the first sub element of this XML element whose tag is equal to the
   *      given tag, or <tt>null</tt> if there is no such sub element.
   */

  public final XMLElement getSubElement (final String tag) {
    if (subElements != null) {
      for (int i = 0; i < subElements.size(); ++i) {
        XMLElement elem = (XMLElement)subElements.get(i);
        if (elem.tag.equals(tag)) {
          return elem;
        }
      }
    }
    return null;
  }

  /**
   * Adds a sub element to this XML element.
   *
   * @param subElement the sub element to be added to this XML element.
   */

  public final void addSubElement (final XMLElement subElement) {
    if (subElements == null) {
      subElements = new ArrayList();
    }
    subElements.add(subElement);
  }

  /**
   * Adds a sub element to this XML element at the given index.
   *
   * @param index the index at which the given element must be added.
   * @param subElement the sub element to be added to this XML element.
   */

  public final void addSubElement (
    final int index,
    final XMLElement subElement)
  {
    if (subElements == null) {
      subElements = new ArrayList();
    }
    subElements.add(index, subElement);
  }

  /**
   * Removes the sub element of this XML element whose index is given.
   *
   * @param index the index of a sub element to be removed.
   */

  public final void removeSubElement (final int index) {
    if (subElements == null) {
      throw new ArrayIndexOutOfBoundsException(index);
    }
    subElements.remove(index);
  }

  // --------------------------------------------------------------------------
  // Basic "manual" validation methods
  // --------------------------------------------------------------------------

  /**
   * Checks the attributes of this element.
   *
   * @param attrs the name of the allowed attributes for this element, separated
   *      by commas (for example "name,value,type").
   * @throws ParserException if an attribute of this element is not specified
   *      in the given allowed attributes list.
   */

  public void checkAttributes (final String attrs) throws ParserException {
    if (attributes != null) {
      String keys = attrs + ',';
      Iterator iter = attributes.keySet().iterator();
      while (iter.hasNext()) {
        String key = (String)iter.next();
        if (keys.indexOf(key + ',') == -1) {
          throw checkErr("\'" + key + "' attribute unsupported");
        }
      }
    }
  }

  /**
   * Checks the sub-elements of this element.
   *
   * @param subElements the allowed sequences of sub element tags for this
   *      element, given as a simplified regular expression: tag | tag* |
   *      tag? | tag,tag.  An example is "title,author,chapter*,annex?".
   * @throws ParserException if the tags of the sub elements of this element do
   *      not conform to the specified regular expression.
   */

  public void checkSubElements (final String subElements)
    throws ParserException
  {
    int index = check(subElements, 0);
    if (this.subElements != null && index < this.subElements.size()) {
      XMLElement subElement = (XMLElement)this.subElements.get(index);
      throw subElement.checkErr("unexpected sub element: " + subElement.tag);
    }
  }

  /**
   * Recursive method to check the sub elements of this element.
   *
   * @param exp a simplified regular expression
   * @param index index of the first sub element not yet checked.
   * @return index of the first sub element that remains to be checked.
   * @throws ParserException if the tags of the sub elements of this element do
   *      not conform to the specified regular expression.
   */

  private int check (final String exp, int index) throws ParserException {
    int comma = exp.indexOf(',');
    if (comma > 0) {
      index = check(exp.substring(0, comma), index);
      return check(exp.substring(comma + 1), index);
    } else if (exp.endsWith("*")) {
      String name = exp.substring(0, exp.length() - 1);
      while (subElements != null && index < subElements.size() &&
             ((XMLElement)subElements.get(index)).tag.equals(name))
      {
        ++index;
      }
      return index;
    } else if (exp.endsWith("?")) {
      String name = exp.substring(0, exp.length() - 1);
      if (subElements != null && index < subElements.size() &&
          ((XMLElement)subElements.get(index)).tag.equals(name))
      {
        ++index;
      }
      return index;
    } else if (exp.length() > 0) {
      if (subElements != null && subElements.size() > index) {
        XMLElement subElement = (XMLElement)subElements.get(index);
        if (!subElement.tag.equals(exp)) {
          throw checkErr(
            "\'" + exp + "' sub element expected at line " + subElement.line);
        }
        return index + 1;
      } else {
        throw checkErr("\'" + exp + "' sub element missing");
      }
    } else {
      return index;
    }
  }

  /**
   * Returns an exception that indicates a validation error.
   *
   * @param msg a detail error message.
   * @return an exception constructed with the given error message.
   */

  private ParserException checkErr (final String msg) {
    return new ParserException(null, line, msg, null);
  }

  // --------------------------------------------------------------------------
  // Printing methods
  // --------------------------------------------------------------------------

  /**
   * Writes the XML element to a writer.
   *
   * @param writer the writer into which this element must be written.
   * @throws IOException if an IO error occurs.
   */

  public void write (final Writer writer) throws IOException {
    write(writer, "", line);
    writer.flush();
  }

  /**
   * Writes the XML element to a writer.
   *
   * @param writer where to write the output to.
   * @param indent how many spaces to indent the element.
   * @param l the current line number in the given writer.
   * @return the new line number in the given writer.
   * @throws IOException if an IO error occurs.
   */

  private int write (
    final Writer writer,
    final String indent,
    int l) throws IOException
  {
    while (l < line) {
      writer.write('\n');
      ++l;
    }
    writer.write(indent);
    if (tag.startsWith("<")) {
      writer.write(tag);
      writer.write('\n');
      return l + 1;
    }
    writer.write('<');
    writer.write(tag);
    if (attributes != null) {
      Iterator iter = attributes.keySet().iterator();
      while (iter.hasNext()) {
        String key = (String)iter.next();
        writer.write(' ');
        writer.write(key);
        writer.write("=\"");
        try {
          writer.write(getAttribute(key, null));
        } catch (ParserException e) {
          // should not happen
        }
        writer.write('"');
      }
    }
    if (subElements != null && subElements.size() > 0) {
      writer.write(">\n");
      ++l;
      String subIndent = indent + "  ";
      for (int i = 0; i < subElements.size(); ++i) {
        l = ((XMLElement)subElements.get(i)).write(writer, subIndent, l);
      }
      writer.write(indent);
      writer.write("</");
      writer.write(tag);
      writer.write(">\n");
    } else {
      writer.write("/>\n");
    }
    return l + 1;
  }

  // --------------------------------------------------------------------------
  // Parsing methods
  // --------------------------------------------------------------------------

  /**
   * Parses an XML element from a {@link LineNumberReader LineNumberReader}.
   *
   * @param reader where to read the XML element.
   * @param parseComment <tt>true</tt> to parse comments, or <tt>false</tt> to
   *      skip them.
   * @return the parsed XML element.
   * @throws IOException if an IO error occurs.
   * @throws ParserException if an XML syntax error is found.
   */

  public static XMLElement parse (
    final LineNumberReader reader,
    final boolean parseComment) throws IOException, ParserException
  {
    r = reader;
    l = reader.getLineNumber();
    try {
      parseChar();
      parseSpaces();
    } catch (EOFException e) {
      return null;
    }
    StringBuffer buf = new StringBuffer();
    while (true) {
      XMLElement e = parse(buf, true, parseComment);
      if (e != null) {
        return e;
      }
    }
  }

  /**
   * Recursive method to parse an XML element.
   *
   * @param buf a buffer that can be used to parse strings.
   * @param parseLT <tt>true</tt> if the first '<' must be parsed, or
   *      <tt>false</tt> if it has already been parsed.
   * @param parseComment <tt>true</tt> to parse comments, or <tt>false</tt> to
   *      skip them.
   * @return the parsed XML element, or <tt>null</tt> if the parsed element is
   *      a comment and <tt>parseComment</tt> is <tt>false</tt>.
   * @throws IOException if an IO error occurs.
   * @throws ParserException if an XML syntax error is found.
   */

  private static XMLElement parse (
    final StringBuffer buf,
    final boolean parseLT,
    final boolean parseComment) throws IOException, ParserException
  {
    if (parseLT) {
      parseSpaces();
      parseChar('<');
    }
    if (c == '!') {
      parseChar();
      if (c == 'D') {
        // skips the <!DOCTYPE ... > definition
        // the following code only works for simple cases of external DTDs
        do {
          parseChar();
        } while (c != '>');
        parseChar();
        return null;
      } else {
        // parses or skips <!-- ... --> comments
        parseChar('-');
        parseChar('-');
        buf.setLength(0);
        buf.append("<!--");
        int dashesToRead = 2;
        while (dashesToRead > 0) {
          dashesToRead = (c == '-' ? dashesToRead - 1 : 2);
          buf.append(c);
          parseChar();
        }
        buf.append(c);
        parseChar('>');
        if (parseComment) {
          return new XMLElement(buf.toString());
        } else {
          return null;
        }
      }
    }
    if (c == '?') {
      // skips the <?xml ... ?> declaration
      do {
        parseChar();
      } while (c != '>');
      parseChar();
      return null;
    }
    XMLElement elt = new XMLElement(parseIdent(buf));
    elt.line = l + 1;
    parseSpaces();
    while (c != '>' && c != '/') {
      String key = parseIdent(buf);
      parseSpaces();
      parseChar('=');
      parseSpaces();
      parseChar('"');
      buf.setLength(0);
      while (c != '"') {
        buf.append(c);
        parseChar();
      }
      parseChar();
      parseSpaces();
      elt.setAttribute(key, buf.toString());
    }
    if (c == '/') {
      parseChar();
      parseChar('>');
      return elt;
    }
    parseChar();
    while (true) {
      parseSpaces();
      parseChar('<');
      if (c == '/') {
        parseChar();
        String s = elt.tag;
        for (int i = 0; i < s.length(); i++) {
          if (c != s.charAt(i)) {
            throw xmlErr("\'" + s + "' expected");
          }
          parseChar();
        }
        if (c != '>') {
          throw xmlErr("\'" + s + "' expected");
        }
        parseChar();
        return elt;
      } else {
        XMLElement e = parse(buf, false, false);
        if (e != null) {
          elt.addSubElement(e);
        }
      }
    }
  }

  /**
   * Parses an identifier.
   *
   * @param buf a buffer that can be used to parse strings.
   * @return the parsed identifier.
   * @throws IOException if an IO error occurs.
   */

  private static String parseIdent (final StringBuffer buf)
    throws IOException
  {
    buf.setLength(0);
    while ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')
           || (c >= '0' && c <= '9') || c == '-' || c == '_')
    {
      buf.append(c);
      parseChar();
    }
    return buf.toString();
  }

  /**
   * Parses a sequence of space characters.
   *
   * @throws IOException if an IO error occurs.
   */

  private static void parseSpaces () throws IOException {
    while (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
      parseChar();
    }
  }

  /**
   * Parses a character.
   *
   * @param expectedChar the expected character.
   * @throws IOException if an IO error occurs.
   * @throws ParserException if the parsed character is not equal to the
   *      expected one.
   */

  private static void parseChar (final char expectedChar)
    throws IOException, ParserException
  {
    if (c != expectedChar) {
      throw xmlErr("\'" + expectedChar + "' expected");
    }
    parseChar();
  }

  /**
   * Parses a character.
   *
   * @throws IOException if an IO error occurs.
   */

  private static void parseChar () throws IOException {
    int i = r.read();
    if (i < 0) {
      throw new EOFException("Unexpected end of file at line " + (l + 1));
    }
    if (i == 10) {
      l = r.getLineNumber();
    }
    c = (char)i;
  }

  /**
   * Returns an exception corresponding to a parsing error.
   *
   * @param msg a detail error message.
   * @return an exception constructed with the given error message.
   */

  private static ParserException xmlErr (final String msg) {
    return new ParserException(null, l, "Syntax error: " + msg, null);
  }
}
