/***
 * FractalGUI: a graphical tool to edit Fractal component configurations.
 * Copyright (C) 2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Authors: Eric Bruneton, Patrice Fauvel
 */

package org.objectweb.fractal.gui.admin.model;

import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.NoSuchInterfaceException;

import org.objectweb.fractal.adl.BasicLoader;
import org.objectweb.fractal.adl.BasicParser;
import org.objectweb.fractal.adl.ResolvingLoader;

import org.objectweb.fractal.util.Fractal;

import org.objectweb.fractal.gui.model.ClientInterface;
import org.objectweb.fractal.gui.model.Component;
import org.objectweb.fractal.gui.model.Configuration;
import org.objectweb.fractal.gui.model.ConfigurationListener;
import org.objectweb.fractal.gui.model.IllegalOperationException;
import org.objectweb.fractal.gui.model.Interface;
import org.objectweb.fractal.gui.model.ServerInterface;
import org.objectweb.fractal.gui.model.VetoableConfigurationListener;
import org.objectweb.fractal.gui.repository.api.Storage;
import org.objectweb.fractal.gui.repository.api.Repository;
import org.objectweb.fractal.gui.admin.model.AdminModelListener;

import java.util.Map;
import java.util.WeakHashMap;
import java.util.List;


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.FontMetrics;



public class BasicAdminModel implements
  BindingController,
  AdminModel,
  VetoableConfigurationListener,
  ConfigurationListener
{

  /**
   * A mandatory client interface bound to a {@link Configuration configuration}
   * model. This is the model on which this model is based and synchronized.
   */

  public final static String CONFIGURATION_BINDING = "configuration";

  /**
   * A mandatory client interface bound to a {@link Storage storage}. This is
   * the storage into which the saved configurations are stored.
   */

  public final static String STORAGE_BINDING = "storage";

  /**
   * A mandatory client interface bound to a {@link Repository repository}. This
   * repository is used to store the configurations in the storage.
   */

  public final static String REPOSITORY_BINDING = "repository";

  /**
   * A mandatory client interface bound to an {@link AdminModelListener}.
   */

  public final static String ADMIN_MODEL_LISTENER_BINDING = "admin-model-listener";

  /**
   * The configuration client interface.
   */

  private Configuration configuration;

  /**
   * The storage client interface.
   */

  private Storage storage;

  /**
   * The repository client interface.
   */

  private Repository repository;

  /**
   * The admin model listener client interface.
   */

  private AdminModelListener adminlistener;

  /**
   * A map associating ComponentIdentity interfaces to Component objects.
   */

  private Map instances;

  private Map compident;

  private int ADD = 0;
  private int REMOVE = 1;
  private int START = 2;
  private int STOP = 3;

  private int range = -1;

  public BasicAdminModel () {
    instances = new WeakHashMap();
    compident = new WeakHashMap();
  }

  // -------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // -------------------------------------------------------------------------

  public String[] listFc () {
    return new String[] {
      CONFIGURATION_BINDING,
      STORAGE_BINDING,
      ADMIN_MODEL_LISTENER_BINDING,
      REPOSITORY_BINDING
    };
  }

  public Object lookupFc (final String clientItfName) {
    if (CONFIGURATION_BINDING.equals(clientItfName)) {
      return configuration;
    } else if (STORAGE_BINDING.equals(clientItfName)) {
      return storage;
    } else if (ADMIN_MODEL_LISTENER_BINDING.equals(clientItfName)) {
      return adminlistener;
    } else if (REPOSITORY_BINDING.equals(clientItfName)) {
      return repository;
    }
    return null;
  }

  public void bindFc (final String clientItfName, final Object serverItf) {
    if (CONFIGURATION_BINDING.equals(clientItfName)) {
      configuration = (Configuration)serverItf;
    } else if (STORAGE_BINDING.equals(clientItfName)) {
      storage = (Storage)serverItf;
    } else if (ADMIN_MODEL_LISTENER_BINDING.equals(clientItfName)) {
      adminlistener = (AdminModelListener)serverItf;
    } else if (REPOSITORY_BINDING.equals(clientItfName)) {
      repository = (Repository)serverItf;
    }
  }

  public void unbindFc (final String clientItfName) {
    if (CONFIGURATION_BINDING.equals(clientItfName)) {
      configuration = null;
    } else if (STORAGE_BINDING.equals(clientItfName)) {
      storage = null;
    } else if (ADMIN_MODEL_LISTENER_BINDING.equals(clientItfName)) {
      adminlistener = null;
    } else if (REPOSITORY_BINDING.equals(clientItfName)) {
      repository = null;
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the AdminModel interface
  // -------------------------------------------------------------------------

  public org.objectweb.fractal.api.Component getInstance (final Component model) {
    if (!configuration.getRootComponent().contains(model)) {
      return null;
    }
    return (org.objectweb.fractal.api.Component)instances.get(model);
  }

  // --- createInstance

  public org.objectweb.fractal.api.Component createInstance (
    final Component model,
    final org.objectweb.fractal.api.Component bootstrapComponent) throws Exception
  {
    if (!configuration.getRootComponent().contains (model)) {
      return null;
    }

    if (instances.size() > 0) {
      throw new IllegalOperationException (
        "Instances already created !");
    }

    // ADL generation
    String execdir = configuration.getExecDir();
    if ((execdir == null) || (execdir.length() < 1)) {
      throw new IllegalOperationException (
        "Exec Directory invalid or missing");
    }

    storage.open ("storage.tmp", true);
    try {
      storage.setFractalFileDir (execdir);
      Component c = configuration.getRootComponent ();
      String root = repository.storeComponent (c);
      storage.setFractalFileDir (null);
    } catch (Exception ex) {
      throw new IllegalOperationException (ex.getMessage());
    } finally {
      storage.close();
    }

    org.objectweb.fractal.api.Component ci = getInstance(model);
    compident.clear();

    if (ci == null) {
      // manual assembly of the parser components
      BasicLoader bl = new BasicLoader ();
      ResolvingLoader rl = new ResolvingLoader ();
      rl.bindFc ("delegate", bl);
      BasicParser bp = new BasicParser();
      bp.bindFc ("loader", rl);
      bp.bindFc ("type-factory", Fractal.getTypeFactory (bootstrapComponent));
      bp.bindFc ("generic-factory", Fractal.getGenericFactory (bootstrapComponent));

      try {
        org.objectweb.fractal.api.Component
          tmpl = bp.loadTemplate (model.getType(), true);
        org.objectweb.fractal.api.Component comp =
          Fractal.getFactory(tmpl).newFcInstance();
        ContentController cc = Fractal.getContentController(comp);

        recurs (comp);
        visite (model, ADD);
      } catch (Exception ex) {
        //ex.printStackTrace(System.out);
      }

      if (instances.size() < 1) {
        throw new IllegalOperationException (
          "No instance created");
      }
      adminlistener.componentCreated(model);
    }
    return ci;
  }

  // --- recurs

  private void recurs (final org.objectweb.fractal.api.Component ci) {
    try {
      String stci = Fractal.getNameController(ci).getFcName();
      compident.put(stci, ci);

      ContentController cc = Fractal.getContentController(ci);
      org.objectweb.fractal.api.Component[] subCi = cc.getFcSubComponents();
      for (int i = 0; i < subCi.length; ++i) {
        org.objectweb.fractal.api.Component sci = subCi[i];
        recurs (sci);
      }
    } catch (Exception ex) {
      /* Il y a une exception systmatique (cf julia.cfg) */
    }
  }

  // --- visite

  private void visite (final Component c, final int action) {
    if (action == ADD) {
      org.objectweb.fractal.api.Component ci =
        (org.objectweb.fractal.api.Component)compident.get(c.getName());
      if (ci == null) {
        // obligatoire pour l'lment de dpart. Je ne sais pas pourquoi.
        ci = (org.objectweb.fractal.api.Component)compident.get(c.getType());
        if (ci == null) return;
      }
      adminlistener.componentCreated(c);
      instances.put(c, ci);
    } else if (action == REMOVE) {
      instances.remove(c);
    }

    List subComponents = c.getSubComponents();
    for (int i = 0; i < subComponents.size(); ++i) {
      Component subC = (Component)subComponents.get(i);
      visite (subC, action);
    }
  }

  // --- deleteInstance

  public void deleteInstance (final Component model) {
    if (!configuration.getRootComponent().contains(model)) {
      return;
    }
    if (this.isStarted(model)) {
      throw new IllegalOperationException(
        "Cannot delete an instance which is started");
    }
    visite (model, REMOVE);
    adminlistener.componentDeleted(model);
  }

  // --- isStarted

  public boolean isStarted (final Component model) {
    org.objectweb.fractal.api.Component ci = getInstance(model);
    if (ci != null) {
      try {
        LifeCycleController lc = Fractal.getLifeCycleController(ci);
        return lc.getFcState().equals("STARTED");
      } catch (NoSuchInterfaceException e) {
        return true;
      }
    }
    return false;
  }

  // --- start

  public void start (final Component model) {
    org.objectweb.fractal.api.Component ci = getInstance(model);

    if (ci != null) {
      try {
        LifeCycleController lc = Fractal.getLifeCycleController(ci);
        if (lc != null) {
          lc.startFc();

          // faire comme runAction ?
          List lser = model.getServerInterfaces();
          ServerInterface ser = (ServerInterface)lser.get(0);
          String signat = ser.getSignature();
          String lien = ser.getName();

          Object iref = null;
          try {
            iref = ci.getFcInterface (lien);
          } catch (Exception ex) {
            ex.printStackTrace();
          }

		      // call main method
		      ClassLoader cl = this.getClass().getClassLoader();
    		  try {
			      Class cla = cl.loadClass(signat);

            java.lang.reflect.Method[] meth = cla.getMethods();
            Object[] arguments = new Object[] { };
            String [] items = new String [meth.length];
            for (int i = 0; i < meth.length; i++) {
              items[i] = meth[i].getName();
            }
            FrameChoiceBox fcb =
              new FrameChoiceBox ("Method for start", items);
            if (range < 0) return;
            meth[range].invoke(iref, arguments);
		      } catch (Exception ex) { ex.printStackTrace(); return; }

          adminlistener.componentStarted(model);
        }
      } catch (NoSuchInterfaceException e) {
      } catch (IllegalLifeCycleException e) {
        throw new IllegalOperationException (
          "Cannot start this instance");
      }
    }
  }

  // --- stop

  public void stop (final Component model) {
    org.objectweb.fractal.api.Component ci = getInstance(model);
    if (ci != null) {
      try {
        LifeCycleController lc = Fractal.getLifeCycleController(ci);
        if (lc != null) {
          lc.stopFc();

          // faire comme runAction ?
          List lser = model.getServerInterfaces();
          ServerInterface ser = (ServerInterface)lser.get(0);
          String signat = ser.getSignature();
          String lien = ser.getName();

          Object iref = null;
          try {
            iref = ci.getFcInterface (lien);
          } catch (Exception ex) {
            ex.printStackTrace();
          }

		      // call main method
		      ClassLoader cl = this.getClass().getClassLoader();
    		  try {
			      Class cla = cl.loadClass(signat);

            java.lang.reflect.Method[] meth = cla.getMethods();
            Object[] arguments = new Object[] { };
            String [] items = new String [meth.length];
            for (int i = 0; i < meth.length; i++) {
              items[i] = meth[i].getName();
            }
            FrameChoiceBox fcb =
              new FrameChoiceBox ("Method for stop", items);
            if (range < 0) return;
            meth[range].invoke(iref, arguments);
		      } catch (Exception ex) { return; }
        }
        adminlistener.componentStopped(model);
      } catch (NoSuchInterfaceException e) {
      } catch (IllegalLifeCycleException e) {
        throw new IllegalOperationException (
          "Cannot stop this instance");
      }
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the VetoableConfigurationListener interface
  // -------------------------------------------------------------------------

  public void canChangeRootComponent () {
    // nothing to do (no veto)
  }

  public void canChangeName (final Component component) {
    // nothing to do (no veto)
  }

  public void canChangeType (final Component component) {
    // nothing to do (no veto)
  }

  public void canChangeImplementation (final Component component) {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot change the implementation of an instanciated component");
    }
  }

  public void canChangeInterfaceName (final Interface i) {
    org.objectweb.fractal.api.Component ci = getInstance(i.getOwner());
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot change the name of an interface of an instanciated component");
    }
  }

  public void canChangeInterfaceSignature (final Interface i) {
    org.objectweb.fractal.api.Component ci = getInstance(i.getOwner());
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot change the signature of an interface of an instanciated component");
    }
  }

  public void canChangeInterfaceContingency (final Interface i) {
    org.objectweb.fractal.api.Component ci = getInstance(i.getOwner());
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot change the contingency of an interface of an instanciated component");
    }
  }

  public void canChangeInterfaceCardinality (final Interface i) {
    org.objectweb.fractal.api.Component ci = getInstance(i.getOwner());
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot change the cardinality of an interface of an instanciated component");
    }
  }

  public void canAddClientInterface (
    final Component component,
    final ClientInterface i)
  {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot add an interface to an instanciated component");
    }
  }

  public void canRemoveClientInterface (
    final Component component,
    final ClientInterface i)
  {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot remove an interface from an instanciated component");
    }
  }

  public void canAddServerInterface (
    final Component component,
    final ServerInterface i)
  {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot add an interface to an instanciated component");
    }
  }

  public void canRemoveServerInterface (
    final Component component,
    final ServerInterface i)
  {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot remove an interface from an instanciated component");
    }
  }

  public void canBindInterface (final ClientInterface citf) {
    if (isStarted(citf.getOwner())) {
      throw new IllegalOperationException(
        "Cannot bind a client interface of a started component");
    }
  }

  public void canRebindInterface (final ClientInterface citf) {
    if (isStarted(citf.getOwner())) {
      throw new IllegalOperationException(
        "Cannot rebind a client interface of a started component");
    }
  }

  public void canUnbindInterface (final ClientInterface citf) {
    if (isStarted(citf.getOwner())) {
      throw new IllegalOperationException(
        "Cannot unbind a client interface of a started component");
    }
  }

  public void canChangeAttributeController (final Component component) {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot change the attribute controller of an instanciated component");
    }
  }

  public void canChangeAttribute (
    final Component component,
    final String attributeName)
  {
    // nothing to do (no veto)
  }

  public void canChangeTemplateControllerDescriptor (
    final Component component)
  {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot change the template controller descriptor of an instanciated component");
    }
  }

  public void canChangeComponentControllerDescriptor (
    final Component component)
  {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      throw new IllegalOperationException(
        "Cannot change the controller descriptor of an instanciated component");
    }
  }

  public void canAddSubComponent (
    final Component parent,
    final Component child)
  {
    if (isStarted(parent) && getInstance(child) != null) {
      throw new IllegalOperationException(
        "Cannot add an instantiated sub component in a started component");
    }
  }

  public void canRemoveSubComponent (
    final Component parent,
    final Component child)
  {
    if (isStarted(parent) && getInstance(child) != null) {
      throw new IllegalOperationException(
        "Cannot remove an instantiated sub component from a started component");
    }
  }

  // -------------------------------------------------------------------------
  // Implementation of the ConfigurationListener interface
  // -------------------------------------------------------------------------

  public void changeCountChanged (final Component component, long changeCount) {
    // does nothing
  }

  public void rootComponentChanged (final Component oldValue) {
    // does nothing
  }

  public void nameChanged (final Component component, final String oldValue) {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      try {
        NameController nc = Fractal.getNameController(ci);
        nc.setFcName(component.getName());
      } catch (NoSuchInterfaceException e) {
      }
    }
  }

  public void typeChanged (final Component component, final String oldValue) {
    // does nothing
  }

  public void implementationChanged (
    final Component component,
    final String oldValue)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void interfaceNameChanged (final Interface i, final String oldValue) {
    // nothing to do (should never be called, because of veto)
  }

  public void interfaceSignatureChanged (
    final Interface i,
    final String oldValue)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void interfaceContingencyChanged (
    final Interface i,
    final boolean oldValue)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void interfaceCardinalityChanged (
    final Interface i,
    final boolean oldValue)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void clientInterfaceAdded (
    final Component component,
    final ClientInterface i,
    final int index)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void clientInterfaceRemoved (
    final Component component,
    final ClientInterface i,
    final int index)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void serverInterfaceAdded (
    final Component component,
    final ServerInterface i,
    final int index)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void serverInterfaceRemoved (
    final Component component,
    final ServerInterface i,
    final int index)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void interfaceBound (
    final ClientInterface citf,
    final ServerInterface sitf)
  {
    org.objectweb.fractal.api.Component ci = getInstance(citf.getOwner());
    if (ci != null) {
      BindingController bc;
      try {
        bc = Fractal.getBindingController(ci);
      } catch (NoSuchInterfaceException e) {
        JOptionPane.showMessageDialog (
          null, e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        return;
      }
      // TODO bindFc on bc
    }
  }

  public void interfaceRebound (
    final ClientInterface citf,
    final ServerInterface oldSitf)
  {
    org.objectweb.fractal.api.Component ci = getInstance(citf.getOwner());
    if (ci != null) {
      BindingController bc;
      try {
       bc = Fractal.getBindingController(ci);
      } catch (NoSuchInterfaceException e) {
        JOptionPane.showMessageDialog (
          null, e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        return;
      }
      // TODO unbindFc, followed by bindFc, on bc
    }
  }

  public void interfaceUnbound (
    final ClientInterface citf,
    final ServerInterface sitf)
  {
    org.objectweb.fractal.api.Component ci = getInstance(citf.getOwner());
    if (ci != null) {
      BindingController bc;
      try {
        bc = Fractal.getBindingController(ci);
      } catch (NoSuchInterfaceException e) {
        JOptionPane.showMessageDialog (
          null, e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        return;
      }
      // TODO unbindFc on bc
    }
  }

  public void attributeControllerChanged (
    final Component component,
    final String oldValue)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void attributeChanged (
    final Component component,
    final String attributeName,
    final String oldValue)
  {
    org.objectweb.fractal.api.Component ci = getInstance(component);
    if (ci != null) {
      // TODO change attribute on ci
    }
  }

  public void templateControllerDescriptorChanged (
    final Component component,
    final String oldValue)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void componentControllerDescriptorChanged (
    final Component component,
    final String oldValue)
  {
    // nothing to do (should never be called, because of veto)
  }

  public void subComponentAdded (
    final Component parent,
    final Component child,
    final int index)
  {
    org.objectweb.fractal.api.Component pi = getInstance(parent);
    org.objectweb.fractal.api.Component ci = getInstance(child);
    if (pi != null && ci != null) {
      ContentController cc;
      try {
        cc = Fractal.getContentController(pi);
      } catch (NoSuchInterfaceException e) {
        JOptionPane.showMessageDialog (
          null, e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        return;
      }
      // TODO addFcSubComponent ci in pi
      // TODO add necessary bindings to or from pi
    }
  }

  public void subComponentRemoved (
    final Component parent,
    final Component child,
    final int index)
  {
    org.objectweb.fractal.api.Component pi = getInstance(parent);
    org.objectweb.fractal.api.Component ci = getInstance(child);
    if (pi != null && ci != null) {
      ContentController cc;
      try {
        cc = Fractal.getContentController(pi);
      } catch (NoSuchInterfaceException e) {
        JOptionPane.showMessageDialog (
          null, e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        return;
      }
      // TODO remove necessary bindings to or from ci
      // TODO removeFcSubComponent ci in pi
    }
  }

/**
 * This subclass asks for a method name.
 */

  /**
   * Font Arial, plain, 13 pts
   */
  static public final Font  fnta13 = new Font ("Arial", Font.PLAIN, 13);

  class FrameChoiceBox extends JDialog implements ActionListener {
	  JPanel panel = new JPanel();
    JButton [] unit;

	  protected void processWindowEvent(WindowEvent e) {
		  if (e.getID() == WindowEvent.WINDOW_CLOSING) { cancel(); }
		  super.processWindowEvent(e);
	  }

	  void cancel() { range = -1; dispose(); }

	  public void actionPerformed (ActionEvent e)	{
      range = -1;
      for (int i = 0; i < unit.length; i++) {
        if (e.getActionCommand().equals(unit[i].getText())) {
          range = i; break;
        }
      }
      dispose();
	  }

	  public FrameChoiceBox (String titre, String [] items)	{
		  unit = new JButton [items.length];
		  enableEvents (AWTEvent.WINDOW_EVENT_MASK);
		  Dimension dim = new Dimension (300, 20+items.length*25);

		  try {
			  this.setTitle(titre);
			  Container cp = getContentPane ();
			  cp.setLayout(new BoxLayout(cp, BoxLayout.Y_AXIS));
			  setResizable(true);

			  Color curcol = Color.lightGray;
			  for (int i = 0; i < items.length; i++) {
				  unit[i] = new JButton (items[i]);
				  unit[i].setPreferredSize(new Dimension(150, 20));
				  unit[i].setFont(fnta13);
				  FontMetrics fm = unit[i].getFontMetrics(fnta13);
				  int d = 0;
				  for (int z = 0; z < items[i].length(); z++)
                d = d + fm.charWidth(items[i].charAt(z));
				  unit[i].setBorder(new EmptyBorder (2, 1, 2, 70-d));
				  unit[i].setBackground(curcol);
				  unit[i].setForeground(Color.black);
				  unit[i].addActionListener(this);
				  cp.add(unit[i]);
			  }
			  setModal(true);
			  this.setSize(dim);
			  this.setLocation(300, 300);
			  pack ();
			  show ();
		  }
		  catch(Exception e) { e.printStackTrace(); }
	  }
  }
// --------------------------
}

